from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.args import feature_view__client_pb2 as _feature_view__client_pb2
from tecton_proto.args import pipeline__client_pb2 as _pipeline__client_pb2
from tecton_proto.args import transformation__client_pb2 as _transformation__client_pb2
from tecton_proto.args import user_defined_function__client_pb2 as _user_defined_function__client_pb2
from tecton_proto.auth import acl__client_pb2 as _acl__client_pb2
from tecton_proto.common import aggregation_function__client_pb2 as _aggregation_function__client_pb2
from tecton_proto.common import data_type__client_pb2 as _data_type__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.common import schema__client_pb2 as _schema__client_pb2
from tecton_proto.common import time_window__client_pb2 as _time_window__client_pb2
from tecton_proto.data import feature_service__client_pb2 as _feature_service__client_pb2
from tecton_proto.data import feature_view__client_pb2 as _feature_view__client_pb2
from tecton_proto.data import realtime_compute__client_pb2 as _realtime_compute__client_pb2
from tecton_proto.data import tecton_api_key__client_pb2 as _tecton_api_key__client_pb2
from tecton_proto.data import transformation__client_pb2 as _transformation__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DATA_TABLE_TIMESTAMP_TYPE_ATTRIBUTE: DataTableTimestampType
DATA_TABLE_TIMESTAMP_TYPE_SORT_KEY: DataTableTimestampType
DATA_TABLE_TIMESTAMP_TYPE_UNKNOWN: DataTableTimestampType
DESCRIPTOR: _descriptor.FileDescriptor
STATUS_TABLE_TIMESTAMP_CONTINUOUS_AGGREGATE: StatusTableTimestampType
STATUS_TABLE_TIMESTAMP_TYPE_ATTRIBUTE: StatusTableTimestampType
STATUS_TABLE_TIMESTAMP_TYPE_SORT_KEY: StatusTableTimestampType
STATUS_TABLE_TIMESTAMP_TYPE_UNKNOWN: StatusTableTimestampType

class CacheGroup(_message.Message):
    __slots__ = ["feature_view_ids", "join_keys", "key_jitter", "key_ttl", "name"]
    FEATURE_VIEW_IDS_FIELD_NUMBER: ClassVar[int]
    JOIN_KEYS_FIELD_NUMBER: ClassVar[int]
    KEY_JITTER_FIELD_NUMBER: ClassVar[int]
    KEY_TTL_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    feature_view_ids: _containers.RepeatedScalarFieldContainer[str]
    join_keys: _containers.RepeatedScalarFieldContainer[str]
    key_jitter: _duration_pb2.Duration
    key_ttl: _duration_pb2.Duration
    name: str
    def __init__(self, name: Optional[str] = ..., join_keys: Optional[Iterable[str]] = ..., key_ttl: Optional[Union[_duration_pb2.Duration, Mapping]] = ..., key_jitter: Optional[Union[_duration_pb2.Duration, Mapping]] = ..., feature_view_ids: Optional[Iterable[str]] = ...) -> None: ...

class CacheParams(_message.Message):
    __slots__ = ["redis"]
    REDIS_FIELD_NUMBER: ClassVar[int]
    redis: _feature_view__client_pb2.RedisOnlineStore
    def __init__(self, redis: Optional[Union[_feature_view__client_pb2.RedisOnlineStore, Mapping]] = ...) -> None: ...

class CanaryConfig(_message.Message):
    __slots__ = ["feature_server_canary_follower_endpoint", "feature_server_canary_id", "feature_server_canary_pod_name"]
    FEATURE_SERVER_CANARY_FOLLOWER_ENDPOINT_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVER_CANARY_ID_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVER_CANARY_POD_NAME_FIELD_NUMBER: ClassVar[int]
    feature_server_canary_follower_endpoint: str
    feature_server_canary_id: str
    feature_server_canary_pod_name: str
    def __init__(self, feature_server_canary_id: Optional[str] = ..., feature_server_canary_pod_name: Optional[str] = ..., feature_server_canary_follower_endpoint: Optional[str] = ...) -> None: ...

class Column(_message.Message):
    __slots__ = ["batch_table_feature_view_index", "data_type", "description", "feature_service_space_name", "feature_view_index", "feature_view_space_name", "tags"]
    class TagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    BATCH_TABLE_FEATURE_VIEW_INDEX_FIELD_NUMBER: ClassVar[int]
    DATA_TYPE_FIELD_NUMBER: ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVICE_SPACE_NAME_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_INDEX_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_SPACE_NAME_FIELD_NUMBER: ClassVar[int]
    TAGS_FIELD_NUMBER: ClassVar[int]
    batch_table_feature_view_index: int
    data_type: _data_type__client_pb2.DataType
    description: str
    feature_service_space_name: str
    feature_view_index: int
    feature_view_space_name: str
    tags: _containers.ScalarMap[str, str]
    def __init__(self, data_type: Optional[Union[_data_type__client_pb2.DataType, Mapping]] = ..., feature_view_space_name: Optional[str] = ..., feature_service_space_name: Optional[str] = ..., feature_view_index: Optional[int] = ..., batch_table_feature_view_index: Optional[int] = ..., description: Optional[str] = ..., tags: Optional[Mapping[str, str]] = ...) -> None: ...

class CompactTransformation(_message.Message):
    __slots__ = ["transformation_id", "transformation_mode", "user_defined_function_id"]
    TRANSFORMATION_ID_FIELD_NUMBER: ClassVar[int]
    TRANSFORMATION_MODE_FIELD_NUMBER: ClassVar[int]
    USER_DEFINED_FUNCTION_ID_FIELD_NUMBER: ClassVar[int]
    transformation_id: _id__client_pb2.Id
    transformation_mode: _transformation__client_pb2.TransformationMode
    user_defined_function_id: str
    def __init__(self, transformation_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., transformation_mode: Optional[Union[_transformation__client_pb2.TransformationMode, str]] = ..., user_defined_function_id: Optional[str] = ...) -> None: ...

class FeaturePlan(_message.Message):
    __slots__ = ["aggregation_function", "aggregation_function_params", "aggregation_leading_edge_mode", "aggregation_secondary_key", "aggregation_window", "batch_table_name", "batch_table_window_index", "cache_index", "data_table_timestamp_type", "deletionTimeWindow", "feature_set_column_hash", "feature_store_format_version", "feature_view_cache_config", "feature_view_id", "feature_view_name", "input_columns", "is_compacted_feature_view", "is_secondary_key_output", "join_keys", "online_store_params", "output_column", "refresh_status_table", "serving_ttl", "slide_period", "status_table_timestamp_type", "stream_table_name", "table_format_version", "table_name", "tiles", "time_window", "timestamp_key", "wildcard_join_keys"]
    AGGREGATION_FUNCTION_FIELD_NUMBER: ClassVar[int]
    AGGREGATION_FUNCTION_PARAMS_FIELD_NUMBER: ClassVar[int]
    AGGREGATION_LEADING_EDGE_MODE_FIELD_NUMBER: ClassVar[int]
    AGGREGATION_SECONDARY_KEY_FIELD_NUMBER: ClassVar[int]
    AGGREGATION_WINDOW_FIELD_NUMBER: ClassVar[int]
    BATCH_TABLE_NAME_FIELD_NUMBER: ClassVar[int]
    BATCH_TABLE_WINDOW_INDEX_FIELD_NUMBER: ClassVar[int]
    CACHE_INDEX_FIELD_NUMBER: ClassVar[int]
    DATA_TABLE_TIMESTAMP_TYPE_FIELD_NUMBER: ClassVar[int]
    DELETIONTIMEWINDOW_FIELD_NUMBER: ClassVar[int]
    FEATURE_SET_COLUMN_HASH_FIELD_NUMBER: ClassVar[int]
    FEATURE_STORE_FORMAT_VERSION_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_CACHE_CONFIG_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_ID_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_NAME_FIELD_NUMBER: ClassVar[int]
    INPUT_COLUMNS_FIELD_NUMBER: ClassVar[int]
    IS_COMPACTED_FEATURE_VIEW_FIELD_NUMBER: ClassVar[int]
    IS_SECONDARY_KEY_OUTPUT_FIELD_NUMBER: ClassVar[int]
    JOIN_KEYS_FIELD_NUMBER: ClassVar[int]
    ONLINE_STORE_PARAMS_FIELD_NUMBER: ClassVar[int]
    OUTPUT_COLUMN_FIELD_NUMBER: ClassVar[int]
    REFRESH_STATUS_TABLE_FIELD_NUMBER: ClassVar[int]
    SERVING_TTL_FIELD_NUMBER: ClassVar[int]
    SLIDE_PERIOD_FIELD_NUMBER: ClassVar[int]
    STATUS_TABLE_TIMESTAMP_TYPE_FIELD_NUMBER: ClassVar[int]
    STREAM_TABLE_NAME_FIELD_NUMBER: ClassVar[int]
    TABLE_FORMAT_VERSION_FIELD_NUMBER: ClassVar[int]
    TABLE_NAME_FIELD_NUMBER: ClassVar[int]
    TILES_FIELD_NUMBER: ClassVar[int]
    TIMESTAMP_KEY_FIELD_NUMBER: ClassVar[int]
    TIME_WINDOW_FIELD_NUMBER: ClassVar[int]
    WILDCARD_JOIN_KEYS_FIELD_NUMBER: ClassVar[int]
    aggregation_function: _aggregation_function__client_pb2.AggregationFunction
    aggregation_function_params: _aggregation_function__client_pb2.AggregationFunctionParams
    aggregation_leading_edge_mode: _feature_view__client_pb2.AggregationLeadingEdge
    aggregation_secondary_key: Column
    aggregation_window: _duration_pb2.Duration
    batch_table_name: str
    batch_table_window_index: int
    cache_index: int
    data_table_timestamp_type: DataTableTimestampType
    deletionTimeWindow: int
    feature_set_column_hash: str
    feature_store_format_version: int
    feature_view_cache_config: _feature_view__client_pb2.FeatureViewCacheConfig
    feature_view_id: str
    feature_view_name: str
    input_columns: _containers.RepeatedCompositeFieldContainer[Column]
    is_compacted_feature_view: bool
    is_secondary_key_output: bool
    join_keys: _containers.RepeatedCompositeFieldContainer[Column]
    online_store_params: _feature_view__client_pb2.OnlineStoreParams
    output_column: Column
    refresh_status_table: bool
    serving_ttl: _duration_pb2.Duration
    slide_period: _duration_pb2.Duration
    status_table_timestamp_type: StatusTableTimestampType
    stream_table_name: str
    table_format_version: int
    table_name: str
    tiles: _containers.RepeatedCompositeFieldContainer[_schema__client_pb2.OnlineBatchTablePartTile]
    time_window: _time_window__client_pb2.TimeWindow
    timestamp_key: str
    wildcard_join_keys: _containers.RepeatedCompositeFieldContainer[Column]
    def __init__(self, output_column: Optional[Union[Column, Mapping]] = ..., input_columns: Optional[Iterable[Union[Column, Mapping]]] = ..., aggregation_function: Optional[Union[_aggregation_function__client_pb2.AggregationFunction, str]] = ..., aggregation_function_params: Optional[Union[_aggregation_function__client_pb2.AggregationFunctionParams, Mapping]] = ..., aggregation_window: Optional[Union[_duration_pb2.Duration, Mapping]] = ..., join_keys: Optional[Iterable[Union[Column, Mapping]]] = ..., wildcard_join_keys: Optional[Iterable[Union[Column, Mapping]]] = ..., aggregation_secondary_key: Optional[Union[Column, Mapping]] = ..., is_secondary_key_output: bool = ..., table_name: Optional[str] = ..., data_table_timestamp_type: Optional[Union[DataTableTimestampType, str]] = ..., status_table_timestamp_type: Optional[Union[StatusTableTimestampType, str]] = ..., timestamp_key: Optional[str] = ..., slide_period: Optional[Union[_duration_pb2.Duration, Mapping]] = ..., serving_ttl: Optional[Union[_duration_pb2.Duration, Mapping]] = ..., refresh_status_table: bool = ..., feature_view_name: Optional[str] = ..., feature_view_id: Optional[str] = ..., feature_store_format_version: Optional[int] = ..., online_store_params: Optional[Union[_feature_view__client_pb2.OnlineStoreParams, Mapping]] = ..., deletionTimeWindow: Optional[int] = ..., time_window: Optional[Union[_time_window__client_pb2.TimeWindow, Mapping]] = ..., feature_view_cache_config: Optional[Union[_feature_view__client_pb2.FeatureViewCacheConfig, Mapping]] = ..., cache_index: Optional[int] = ..., table_format_version: Optional[int] = ..., batch_table_name: Optional[str] = ..., batch_table_window_index: Optional[int] = ..., stream_table_name: Optional[str] = ..., tiles: Optional[Iterable[Union[_schema__client_pb2.OnlineBatchTablePartTile, Mapping]]] = ..., is_compacted_feature_view: bool = ..., feature_set_column_hash: Optional[str] = ..., aggregation_leading_edge_mode: Optional[Union[_feature_view__client_pb2.AggregationLeadingEdge, str]] = ...) -> None: ...

class FeatureServerConfiguration(_message.Message):
    __slots__ = ["all_online_compute_configs", "all_online_store_params", "authorized_api_keys", "cache_groups", "computed_time", "feature_server_canary_config", "feature_service_acls", "feature_services", "global_table_config_by_name", "jwks", "remote_compute_configs", "user_defined_function_map", "workspace_acls"]
    class CacheGroupsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: CacheGroup
        def __init__(self, key: Optional[str] = ..., value: Optional[Union[CacheGroup, Mapping]] = ...) -> None: ...
    class GlobalTableConfigByNameEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: GlobalTableConfig
        def __init__(self, key: Optional[str] = ..., value: Optional[Union[GlobalTableConfig, Mapping]] = ...) -> None: ...
    class JwksEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: Jwk
        def __init__(self, key: Optional[str] = ..., value: Optional[Union[Jwk, Mapping]] = ...) -> None: ...
    class UserDefinedFunctionMapEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: _user_defined_function__client_pb2.UserDefinedFunction
        def __init__(self, key: Optional[str] = ..., value: Optional[Union[_user_defined_function__client_pb2.UserDefinedFunction, Mapping]] = ...) -> None: ...
    ALL_ONLINE_COMPUTE_CONFIGS_FIELD_NUMBER: ClassVar[int]
    ALL_ONLINE_STORE_PARAMS_FIELD_NUMBER: ClassVar[int]
    AUTHORIZED_API_KEYS_FIELD_NUMBER: ClassVar[int]
    CACHE_GROUPS_FIELD_NUMBER: ClassVar[int]
    COMPUTED_TIME_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVER_CANARY_CONFIG_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVICES_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVICE_ACLS_FIELD_NUMBER: ClassVar[int]
    GLOBAL_TABLE_CONFIG_BY_NAME_FIELD_NUMBER: ClassVar[int]
    JWKS_FIELD_NUMBER: ClassVar[int]
    REMOTE_COMPUTE_CONFIGS_FIELD_NUMBER: ClassVar[int]
    USER_DEFINED_FUNCTION_MAP_FIELD_NUMBER: ClassVar[int]
    WORKSPACE_ACLS_FIELD_NUMBER: ClassVar[int]
    all_online_compute_configs: _containers.RepeatedCompositeFieldContainer[_realtime_compute__client_pb2.OnlineComputeConfig]
    all_online_store_params: _containers.RepeatedCompositeFieldContainer[_feature_view__client_pb2.OnlineStoreParams]
    authorized_api_keys: _containers.RepeatedCompositeFieldContainer[_tecton_api_key__client_pb2.TectonApiKey]
    cache_groups: _containers.MessageMap[str, CacheGroup]
    computed_time: _timestamp_pb2.Timestamp
    feature_server_canary_config: CanaryConfig
    feature_service_acls: _containers.RepeatedCompositeFieldContainer[FeatureServiceAcls]
    feature_services: _containers.RepeatedCompositeFieldContainer[FeatureServicePlan]
    global_table_config_by_name: _containers.MessageMap[str, GlobalTableConfig]
    jwks: _containers.MessageMap[str, Jwk]
    remote_compute_configs: _containers.RepeatedCompositeFieldContainer[_realtime_compute__client_pb2.RemoteFunctionComputeConfig]
    user_defined_function_map: _containers.MessageMap[str, _user_defined_function__client_pb2.UserDefinedFunction]
    workspace_acls: _containers.RepeatedCompositeFieldContainer[WorkspaceAcls]
    def __init__(self, computed_time: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., feature_services: Optional[Iterable[Union[FeatureServicePlan, Mapping]]] = ..., global_table_config_by_name: Optional[Mapping[str, GlobalTableConfig]] = ..., authorized_api_keys: Optional[Iterable[Union[_tecton_api_key__client_pb2.TectonApiKey, Mapping]]] = ..., feature_service_acls: Optional[Iterable[Union[FeatureServiceAcls, Mapping]]] = ..., workspace_acls: Optional[Iterable[Union[WorkspaceAcls, Mapping]]] = ..., all_online_store_params: Optional[Iterable[Union[_feature_view__client_pb2.OnlineStoreParams, Mapping]]] = ..., feature_server_canary_config: Optional[Union[CanaryConfig, Mapping]] = ..., remote_compute_configs: Optional[Iterable[Union[_realtime_compute__client_pb2.RemoteFunctionComputeConfig, Mapping]]] = ..., all_online_compute_configs: Optional[Iterable[Union[_realtime_compute__client_pb2.OnlineComputeConfig, Mapping]]] = ..., cache_groups: Optional[Mapping[str, CacheGroup]] = ..., user_defined_function_map: Optional[Mapping[str, _user_defined_function__client_pb2.UserDefinedFunction]] = ..., jwks: Optional[Mapping[str, Jwk]] = ...) -> None: ...

class FeatureServiceAcls(_message.Message):
    __slots__ = ["acls", "feature_service_id"]
    ACLS_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVICE_ID_FIELD_NUMBER: ClassVar[int]
    acls: _containers.RepeatedCompositeFieldContainer[_acl__client_pb2.Acl]
    feature_service_id: _id__client_pb2.Id
    def __init__(self, feature_service_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., acls: Optional[Iterable[Union[_acl__client_pb2.Acl, Mapping]]] = ...) -> None: ...

class FeatureServiceCachePlan(_message.Message):
    __slots__ = ["cache_group_name", "feature_set_column_hashes", "feature_view_ids", "remapped_join_key_lists"]
    CACHE_GROUP_NAME_FIELD_NUMBER: ClassVar[int]
    FEATURE_SET_COLUMN_HASHES_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_IDS_FIELD_NUMBER: ClassVar[int]
    REMAPPED_JOIN_KEY_LISTS_FIELD_NUMBER: ClassVar[int]
    cache_group_name: str
    feature_set_column_hashes: _containers.RepeatedScalarFieldContainer[str]
    feature_view_ids: _containers.RepeatedScalarFieldContainer[str]
    remapped_join_key_lists: _containers.RepeatedCompositeFieldContainer[RemappedJoinKeys]
    def __init__(self, feature_view_ids: Optional[Iterable[str]] = ..., cache_group_name: Optional[str] = ..., remapped_join_key_lists: Optional[Iterable[Union[RemappedJoinKeys, Mapping]]] = ..., feature_set_column_hashes: Optional[Iterable[str]] = ...) -> None: ...

class FeatureServicePlan(_message.Message):
    __slots__ = ["cache_plans", "feature_service_id", "feature_service_name", "feature_view_id", "feature_view_name", "features_plans", "join_key_template", "logging_config", "realtime_environment", "workspace_name", "workspace_state_id"]
    CACHE_PLANS_FIELD_NUMBER: ClassVar[int]
    FEATURES_PLANS_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVICE_ID_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVICE_NAME_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_ID_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_NAME_FIELD_NUMBER: ClassVar[int]
    JOIN_KEY_TEMPLATE_FIELD_NUMBER: ClassVar[int]
    LOGGING_CONFIG_FIELD_NUMBER: ClassVar[int]
    REALTIME_ENVIRONMENT_FIELD_NUMBER: ClassVar[int]
    WORKSPACE_NAME_FIELD_NUMBER: ClassVar[int]
    WORKSPACE_STATE_ID_FIELD_NUMBER: ClassVar[int]
    cache_plans: _containers.RepeatedCompositeFieldContainer[FeatureServiceCachePlan]
    feature_service_id: _id__client_pb2.Id
    feature_service_name: str
    feature_view_id: _id__client_pb2.Id
    feature_view_name: str
    features_plans: _containers.RepeatedCompositeFieldContainer[FeaturesPlan]
    join_key_template: _feature_service__client_pb2.JoinKeyTemplate
    logging_config: LoggingConfig
    realtime_environment: _realtime_compute__client_pb2.OnlineComputeConfig
    workspace_name: str
    workspace_state_id: _id__client_pb2.Id
    def __init__(self, feature_service_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., feature_view_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., feature_service_name: Optional[str] = ..., feature_view_name: Optional[str] = ..., workspace_name: Optional[str] = ..., workspace_state_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., features_plans: Optional[Iterable[Union[FeaturesPlan, Mapping]]] = ..., join_key_template: Optional[Union[_feature_service__client_pb2.JoinKeyTemplate, Mapping]] = ..., logging_config: Optional[Union[LoggingConfig, Mapping]] = ..., realtime_environment: Optional[Union[_realtime_compute__client_pb2.OnlineComputeConfig, Mapping]] = ..., cache_plans: Optional[Iterable[Union[FeatureServiceCachePlan, Mapping]]] = ...) -> None: ...

class FeatureVectorPlan(_message.Message):
    __slots__ = ["features"]
    FEATURES_FIELD_NUMBER: ClassVar[int]
    features: _containers.RepeatedCompositeFieldContainer[FeaturePlan]
    def __init__(self, features: Optional[Iterable[Union[FeaturePlan, Mapping]]] = ...) -> None: ...

class FeaturesPlan(_message.Message):
    __slots__ = ["feature_plan", "realtime_features_plan"]
    FEATURE_PLAN_FIELD_NUMBER: ClassVar[int]
    REALTIME_FEATURES_PLAN_FIELD_NUMBER: ClassVar[int]
    feature_plan: FeaturePlan
    realtime_features_plan: RealtimeFeaturesPlan
    def __init__(self, feature_plan: Optional[Union[FeaturePlan, Mapping]] = ..., realtime_features_plan: Optional[Union[RealtimeFeaturesPlan, Mapping]] = ...) -> None: ...

class GlobalTableConfig(_message.Message):
    __slots__ = ["feature_data_water_mark", "feature_store_format_version", "feature_view_id", "feature_view_name", "online_store_params", "refresh_status_table", "slide_period", "status_table_timestamp_type", "table_format_version", "workspace_name"]
    FEATURE_DATA_WATER_MARK_FIELD_NUMBER: ClassVar[int]
    FEATURE_STORE_FORMAT_VERSION_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_ID_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_NAME_FIELD_NUMBER: ClassVar[int]
    ONLINE_STORE_PARAMS_FIELD_NUMBER: ClassVar[int]
    REFRESH_STATUS_TABLE_FIELD_NUMBER: ClassVar[int]
    SLIDE_PERIOD_FIELD_NUMBER: ClassVar[int]
    STATUS_TABLE_TIMESTAMP_TYPE_FIELD_NUMBER: ClassVar[int]
    TABLE_FORMAT_VERSION_FIELD_NUMBER: ClassVar[int]
    WORKSPACE_NAME_FIELD_NUMBER: ClassVar[int]
    feature_data_water_mark: _timestamp_pb2.Timestamp
    feature_store_format_version: int
    feature_view_id: _id__client_pb2.Id
    feature_view_name: str
    online_store_params: _feature_view__client_pb2.OnlineStoreParams
    refresh_status_table: bool
    slide_period: _duration_pb2.Duration
    status_table_timestamp_type: StatusTableTimestampType
    table_format_version: int
    workspace_name: str
    def __init__(self, feature_view_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., feature_view_name: Optional[str] = ..., workspace_name: Optional[str] = ..., slide_period: Optional[Union[_duration_pb2.Duration, Mapping]] = ..., status_table_timestamp_type: Optional[Union[StatusTableTimestampType, str]] = ..., refresh_status_table: bool = ..., feature_store_format_version: Optional[int] = ..., online_store_params: Optional[Union[_feature_view__client_pb2.OnlineStoreParams, Mapping]] = ..., table_format_version: Optional[int] = ..., feature_data_water_mark: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ...) -> None: ...

class Jwk(_message.Message):
    __slots__ = ["alg", "n", "use"]
    ALG_FIELD_NUMBER: ClassVar[int]
    N_FIELD_NUMBER: ClassVar[int]
    USE_FIELD_NUMBER: ClassVar[int]
    alg: str
    n: str
    use: str
    def __init__(self, n: Optional[str] = ..., alg: Optional[str] = ..., use: Optional[str] = ...) -> None: ...

class LoggingConfig(_message.Message):
    __slots__ = ["avro_schema", "log_effective_times", "sample_rate"]
    AVRO_SCHEMA_FIELD_NUMBER: ClassVar[int]
    LOG_EFFECTIVE_TIMES_FIELD_NUMBER: ClassVar[int]
    SAMPLE_RATE_FIELD_NUMBER: ClassVar[int]
    avro_schema: str
    log_effective_times: bool
    sample_rate: float
    def __init__(self, sample_rate: Optional[float] = ..., log_effective_times: bool = ..., avro_schema: Optional[str] = ...) -> None: ...

class RealtimeFeaturesPlan(_message.Message):
    __slots__ = ["args_from_request_context", "compact_transformations", "description", "feature_set_inputs", "feature_view_id", "feature_view_name", "outputs", "pipeline", "tags", "transformations"]
    class FeatureSetInputsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: FeatureVectorPlan
        def __init__(self, key: Optional[str] = ..., value: Optional[Union[FeatureVectorPlan, Mapping]] = ...) -> None: ...
    class TagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    ARGS_FROM_REQUEST_CONTEXT_FIELD_NUMBER: ClassVar[int]
    COMPACT_TRANSFORMATIONS_FIELD_NUMBER: ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: ClassVar[int]
    FEATURE_SET_INPUTS_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_ID_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_NAME_FIELD_NUMBER: ClassVar[int]
    OUTPUTS_FIELD_NUMBER: ClassVar[int]
    PIPELINE_FIELD_NUMBER: ClassVar[int]
    TAGS_FIELD_NUMBER: ClassVar[int]
    TRANSFORMATIONS_FIELD_NUMBER: ClassVar[int]
    args_from_request_context: _containers.RepeatedCompositeFieldContainer[Column]
    compact_transformations: _containers.RepeatedCompositeFieldContainer[CompactTransformation]
    description: str
    feature_set_inputs: _containers.MessageMap[str, FeatureVectorPlan]
    feature_view_id: str
    feature_view_name: str
    outputs: _containers.RepeatedCompositeFieldContainer[Column]
    pipeline: _pipeline__client_pb2.Pipeline
    tags: _containers.ScalarMap[str, str]
    transformations: _containers.RepeatedCompositeFieldContainer[_transformation__client_pb2.Transformation]
    def __init__(self, args_from_request_context: Optional[Iterable[Union[Column, Mapping]]] = ..., outputs: Optional[Iterable[Union[Column, Mapping]]] = ..., feature_set_inputs: Optional[Mapping[str, FeatureVectorPlan]] = ..., pipeline: Optional[Union[_pipeline__client_pb2.Pipeline, Mapping]] = ..., transformations: Optional[Iterable[Union[_transformation__client_pb2.Transformation, Mapping]]] = ..., feature_view_name: Optional[str] = ..., feature_view_id: Optional[str] = ..., compact_transformations: Optional[Iterable[Union[CompactTransformation, Mapping]]] = ..., description: Optional[str] = ..., tags: Optional[Mapping[str, str]] = ...) -> None: ...

class RemappedJoinKeys(_message.Message):
    __slots__ = ["join_keys"]
    JOIN_KEYS_FIELD_NUMBER: ClassVar[int]
    join_keys: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, join_keys: Optional[Iterable[str]] = ...) -> None: ...

class WorkspaceAcls(_message.Message):
    __slots__ = ["acls", "workspace_name"]
    ACLS_FIELD_NUMBER: ClassVar[int]
    WORKSPACE_NAME_FIELD_NUMBER: ClassVar[int]
    acls: _containers.RepeatedCompositeFieldContainer[_acl__client_pb2.Acl]
    workspace_name: str
    def __init__(self, workspace_name: Optional[str] = ..., acls: Optional[Iterable[Union[_acl__client_pb2.Acl, Mapping]]] = ...) -> None: ...

class DataTableTimestampType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class StatusTableTimestampType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from tecton_proto.args import entity__client_pb2 as _entity__client_pb2
from tecton_proto.args import feature_service__client_pb2 as _feature_service__client_pb2
from tecton_proto.args import feature_view__client_pb2 as _feature_view__client_pb2
from tecton_proto.args import server_group__client_pb2 as _server_group__client_pb2
from tecton_proto.args import transformation__client_pb2 as _transformation__client_pb2
from tecton_proto.args import virtual_data_source__client_pb2 as _virtual_data_source__client_pb2
from tecton_proto.common import schema__client_pb2 as _schema__client_pb2
from tecton_proto.common import spark_schema__client_pb2 as _spark_schema__client_pb2
from tecton_proto.modelartifactservice import model_artifact_service__client_pb2 as _model_artifact_service__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class EntityValidationArgs(_message.Message):
    __slots__ = ["args"]
    ARGS_FIELD_NUMBER: ClassVar[int]
    args: _entity__client_pb2.EntityArgs
    def __init__(self, args: Optional[Union[_entity__client_pb2.EntityArgs, Mapping]] = ...) -> None: ...

class FcoValidationArgs(_message.Message):
    __slots__ = ["entity", "feature_service", "feature_view", "server_group", "transformation", "virtual_data_source"]
    ENTITY_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVICE_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_FIELD_NUMBER: ClassVar[int]
    SERVER_GROUP_FIELD_NUMBER: ClassVar[int]
    TRANSFORMATION_FIELD_NUMBER: ClassVar[int]
    VIRTUAL_DATA_SOURCE_FIELD_NUMBER: ClassVar[int]
    entity: EntityValidationArgs
    feature_service: FeatureServiceValidationArgs
    feature_view: FeatureViewValidationArgs
    server_group: ServerGroupValidationArgs
    transformation: TransformationValidationArgs
    virtual_data_source: VirtualDataSourceValidationArgs
    def __init__(self, virtual_data_source: Optional[Union[VirtualDataSourceValidationArgs, Mapping]] = ..., entity: Optional[Union[EntityValidationArgs, Mapping]] = ..., feature_view: Optional[Union[FeatureViewValidationArgs, Mapping]] = ..., feature_service: Optional[Union[FeatureServiceValidationArgs, Mapping]] = ..., transformation: Optional[Union[TransformationValidationArgs, Mapping]] = ..., server_group: Optional[Union[ServerGroupValidationArgs, Mapping]] = ...) -> None: ...

class FeatureServiceValidationArgs(_message.Message):
    __slots__ = ["args"]
    ARGS_FIELD_NUMBER: ClassVar[int]
    args: _feature_service__client_pb2.FeatureServiceArgs
    def __init__(self, args: Optional[Union[_feature_service__client_pb2.FeatureServiceArgs, Mapping]] = ...) -> None: ...

class FeatureViewValidationArgs(_message.Message):
    __slots__ = ["args", "local_model_artifacts", "materialization_schema", "view_schema"]
    ARGS_FIELD_NUMBER: ClassVar[int]
    LOCAL_MODEL_ARTIFACTS_FIELD_NUMBER: ClassVar[int]
    MATERIALIZATION_SCHEMA_FIELD_NUMBER: ClassVar[int]
    VIEW_SCHEMA_FIELD_NUMBER: ClassVar[int]
    args: _feature_view__client_pb2.FeatureViewArgs
    local_model_artifacts: _containers.RepeatedCompositeFieldContainer[_model_artifact_service__client_pb2.ModelArtifactInfo]
    materialization_schema: _schema__client_pb2.Schema
    view_schema: _schema__client_pb2.Schema
    def __init__(self, args: Optional[Union[_feature_view__client_pb2.FeatureViewArgs, Mapping]] = ..., view_schema: Optional[Union[_schema__client_pb2.Schema, Mapping]] = ..., materialization_schema: Optional[Union[_schema__client_pb2.Schema, Mapping]] = ..., local_model_artifacts: Optional[Iterable[Union[_model_artifact_service__client_pb2.ModelArtifactInfo, Mapping]]] = ...) -> None: ...

class ServerGroupValidationArgs(_message.Message):
    __slots__ = ["args"]
    ARGS_FIELD_NUMBER: ClassVar[int]
    args: _server_group__client_pb2.ServerGroupArgs
    def __init__(self, args: Optional[Union[_server_group__client_pb2.ServerGroupArgs, Mapping]] = ...) -> None: ...

class TransformationValidationArgs(_message.Message):
    __slots__ = ["args"]
    ARGS_FIELD_NUMBER: ClassVar[int]
    args: _transformation__client_pb2.TransformationArgs
    def __init__(self, args: Optional[Union[_transformation__client_pb2.TransformationArgs, Mapping]] = ...) -> None: ...

class ValidationRequest(_message.Message):
    __slots__ = ["validation_args"]
    VALIDATION_ARGS_FIELD_NUMBER: ClassVar[int]
    validation_args: _containers.RepeatedCompositeFieldContainer[FcoValidationArgs]
    def __init__(self, validation_args: Optional[Iterable[Union[FcoValidationArgs, Mapping]]] = ...) -> None: ...

class VirtualDataSourceValidationArgs(_message.Message):
    __slots__ = ["args", "batch_schema", "stream_schema"]
    ARGS_FIELD_NUMBER: ClassVar[int]
    BATCH_SCHEMA_FIELD_NUMBER: ClassVar[int]
    STREAM_SCHEMA_FIELD_NUMBER: ClassVar[int]
    args: _virtual_data_source__client_pb2.VirtualDataSourceArgs
    batch_schema: _spark_schema__client_pb2.SparkSchema
    stream_schema: _spark_schema__client_pb2.SparkSchema
    def __init__(self, args: Optional[Union[_virtual_data_source__client_pb2.VirtualDataSourceArgs, Mapping]] = ..., batch_schema: Optional[Union[_spark_schema__client_pb2.SparkSchema, Mapping]] = ..., stream_schema: Optional[Union[_spark_schema__client_pb2.SparkSchema, Mapping]] = ...) -> None: ...



BUILD_STATUS = {'BUILD_ID': 'uncommitted-buildkite-agent-1730316131', 'BUILD_TIMESTAMP': '1730316131', 'FORMATTED_DATE': '2024 Oct 30 19 22 11 Wed', 'GIT_COMMIT': '0289549850ef55f040ea02da5144e54687aeef58', 'BUILD_EMBED_LABEL': '', 'BUILD_HOST': 'buildkite-agent-medium-6cf65dcd96-66qd6', 'BUILD_USER': 'buildkite-agent'}



"""
Tecton's sdk currently relies on the pendulum package for datetime functionality. In order for the sdk to support python
3.12, we must upgrade pendulum to 3.X. However, Tecton must also support python 3.7 at this time, which pendulum 3.X
does not support.

To make the sdk support python 3.7 - 3.12, this compat module mediates access to pendulum functionality: pendulum 3.0.X
for python>=3.12, and pendulum 2.1.2 for python<3.12 (as specified in sdk/pypi/pyproject.toml). The eventual goal is
removing the pendulum dependency entirely.
"""

import re

import pendulum as _pendulum


_VERSION_3_0_PATTERN = r"^3\.0\.\d+$"
_IS_PENDULUM_3 = hasattr(_pendulum, "__version__") and re.match(_VERSION_3_0_PATTERN, getattr(_pendulum, "__version__"))

# classes
Duration = _pendulum.Duration

# modules
tz = _pendulum.tz

# functions
now = _pendulum.now
from_timestamp = _pendulum.from_timestamp
datetime = _pendulum.datetime
duration = _pendulum.duration
instance = _pendulum.instance
parse = _pendulum.parse
today = _pendulum.today
timezone = _pendulum.timezone

if _IS_PENDULUM_3:
    """
    In pendulum 2, DateTime.__str__ returned iso8601 based timestamps e.g. '1975-05-21T22:00:00'. However,
    in pendulum 3, DateTime.__str__ returns timestamps without the 'T' to match the native python datetime behavior.
    tecton_athena sql templates like sdk/tecton_athena/templates/time_limit.sql explicitly expect iso8601 timestamps
    and tests assert the 'T' is present. In order to maintain compatibility with our code, we override __str__ here.
    """

    def pendulum_V2_str(self):
        return self.isoformat("T")

    DateTime = _pendulum.DateTime
    DateTime.__str__ = pendulum_V2_str

    """
    In pendulum 3, the Period class and period method have been renamed to Interval and interval, respectively.
    Rename them to Period for compatibility.
    """
    Period = _pendulum.Interval
    period = _pendulum.interval

else:
    DateTime = _pendulum.DateTime
    Period = _pendulum.Period
    period = _pendulum.period



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/server_group__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import container_image__client_pb2 as tecton__proto_dot_common_dot_container__image____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.common import scaling_config__client_pb2 as tecton__proto_dot_common_dot_scaling__config____client__pb2
from tecton_proto.common import server_group_type__client_pb2 as tecton__proto_dot_common_dot_server__group__type____client__pb2
from tecton_proto.data import fco_metadata__client_pb2 as tecton__proto_dot_data_dot_fco__metadata____client__pb2
from tecton_proto.validation import validator__client_pb2 as tecton__proto_dot_validation_dot_validator____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,tecton_proto/data/server_group__client.proto\x12\x11tecton_proto.data\x1a\x31tecton_proto/common/container_image__client.proto\x1a$tecton_proto/common/id__client.proto\x1a\x30tecton_proto/common/scaling_config__client.proto\x1a\x33tecton_proto/common/server_group_type__client.proto\x1a,tecton_proto/data/fco_metadata__client.proto\x1a/tecton_proto/validation/validator__client.proto\"\xee\x06\n\x0bServerGroup\x12?\n\x0fserver_group_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rserverGroupId\x12\x41\n\x0c\x66\x63o_metadata\x18\x02 \x01(\x0b\x32\x1e.tecton_proto.data.FcoMetadataR\x0b\x66\x63oMetadata\x12\x38\n\x04type\x18\x03 \x01(\x0e\x32$.tecton_proto.common.ServerGroupTypeR\x04type\x12_\n\x16transform_server_group\x18\x08 \x01(\x0b\x32\'.tecton_proto.data.TransformServerGroupH\x00R\x14transformServerGroup\x12Y\n\x14\x66\x65\x61ture_server_group\x18\n \x01(\x0b\x32%.tecton_proto.data.FeatureServerGroupH\x00R\x12\x66\x65\x61tureServerGroup\x12\x45\n\x07options\x18\t \x03(\x0b\x32+.tecton_proto.data.ServerGroup.OptionsEntryR\x07options\x12[\n\x0fvalidation_args\x18\x0b \x01(\x0b\x32\x32.tecton_proto.validation.ServerGroupValidationArgsR\x0evalidationArgs\x12W\n\x12\x61utoscaling_config\x18\x0c \x01(\x0b\x32&.tecton_proto.common.AutoscalingConfigH\x01R\x11\x61utoscalingConfig\x12m\n\x1aprovisioned_scaling_config\x18\r \x01(\x0b\x32-.tecton_proto.common.ProvisionedScalingConfigH\x01R\x18provisionedScalingConfig\x1a:\n\x0cOptionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x13\n\x11server_group_typeB\x10\n\x0escaling_configJ\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06J\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08\"\xac\x01\n\x14TransformServerGroup\x12%\n\x0e\x65nvironment_id\x18\x01 \x01(\tR\renvironmentId\x12)\n\x10\x65nvironment_name\x18\x02 \x01(\tR\x0f\x65nvironmentName\x12\x42\n\nimage_info\x18\x03 \x01(\x0b\x32#.tecton_proto.common.ContainerImageR\timageInfo\"\x14\n\x12\x46\x65\x61tureServerGroupB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.server_group__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _SERVERGROUP_OPTIONSENTRY._options = None
  _SERVERGROUP_OPTIONSENTRY._serialized_options = b'8\001'
  _SERVERGROUP._serialized_start=355
  _SERVERGROUP._serialized_end=1233
  _SERVERGROUP_OPTIONSENTRY._serialized_start=1112
  _SERVERGROUP_OPTIONSENTRY._serialized_end=1170
  _TRANSFORMSERVERGROUP._serialized_start=1236
  _TRANSFORMSERVERGROUP._serialized_end=1408
  _FEATURESERVERGROUP._serialized_start=1410
  _FEATURESERVERGROUP._serialized_end=1430
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar

COPIER_STATUS_ERROR: OnlineStoreCopierTaskState
COPIER_STATUS_PENDING: OnlineStoreCopierTaskState
COPIER_STATUS_RUNNING: OnlineStoreCopierTaskState
COPIER_STATUS_SUCCESS: OnlineStoreCopierTaskState
COPIER_STATUS_UNSPECIFIED: OnlineStoreCopierTaskState
DESCRIPTOR: _descriptor.FileDescriptor
MATERIALIZATION_ATTEMPT_STATUS_CANCELLED: MaterializationTaskAttemptState
MATERIALIZATION_ATTEMPT_STATUS_ERROR: MaterializationTaskAttemptState
MATERIALIZATION_ATTEMPT_STATUS_PENDING: MaterializationTaskAttemptState
MATERIALIZATION_ATTEMPT_STATUS_RUNNING: MaterializationTaskAttemptState
MATERIALIZATION_ATTEMPT_STATUS_SUCCESS: MaterializationTaskAttemptState
MATERIALIZATION_ATTEMPT_STATUS_UNSPECIFIED: MaterializationTaskAttemptState
MATERIALIZATION_TASK_STATUS_DRAINED: MaterializationTaskState
MATERIALIZATION_TASK_STATUS_DRAINING: MaterializationTaskState
MATERIALIZATION_TASK_STATUS_MANUALLY_CANCELLED: MaterializationTaskState
MATERIALIZATION_TASK_STATUS_MANUAL_CANCELLATION_REQUESTED: MaterializationTaskState
MATERIALIZATION_TASK_STATUS_MANUAL_RETRY: MaterializationTaskState
MATERIALIZATION_TASK_STATUS_PERMANENT_FAILURE: MaterializationTaskState
MATERIALIZATION_TASK_STATUS_RUNNING: MaterializationTaskState
MATERIALIZATION_TASK_STATUS_SUCCESS: MaterializationTaskState
MATERIALIZATION_TASK_STATUS_UNSPECIFIED: MaterializationTaskState

class MaterializationTaskState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MaterializationTaskAttemptState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class OnlineStoreCopierTaskState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from tecton_proto.args import basic_info__client_pb2 as _basic_info__client_pb2
from tecton_proto.args import diff_options__client_pb2 as _diff_options__client_pb2
from tecton_proto.common import framework_version__client_pb2 as _framework_version__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class ColumnPair(_message.Message):
    __slots__ = ["feature_column", "spine_column"]
    FEATURE_COLUMN_FIELD_NUMBER: ClassVar[int]
    SPINE_COLUMN_FIELD_NUMBER: ClassVar[int]
    feature_column: str
    spine_column: str
    def __init__(self, spine_column: Optional[str] = ..., feature_column: Optional[str] = ...) -> None: ...

class FeatureReference(_message.Message):
    __slots__ = ["feature_view_id", "features", "namespace", "override_join_keys"]
    FEATURES_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_ID_FIELD_NUMBER: ClassVar[int]
    NAMESPACE_FIELD_NUMBER: ClassVar[int]
    OVERRIDE_JOIN_KEYS_FIELD_NUMBER: ClassVar[int]
    feature_view_id: _id__client_pb2.Id
    features: _containers.RepeatedScalarFieldContainer[str]
    namespace: str
    override_join_keys: _containers.RepeatedCompositeFieldContainer[ColumnPair]
    def __init__(self, feature_view_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., override_join_keys: Optional[Iterable[Union[ColumnPair, Mapping]]] = ..., namespace: Optional[str] = ..., features: Optional[Iterable[str]] = ...) -> None: ...

class FeatureServiceArgs(_message.Message):
    __slots__ = ["enable_online_caching", "feature_references", "feature_server_group", "feature_service_id", "info", "logging", "online_serving_enabled", "options", "prevent_destroy", "realtime_environment", "transform_server_group", "version"]
    class OptionsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    ENABLE_ONLINE_CACHING_FIELD_NUMBER: ClassVar[int]
    FEATURE_REFERENCES_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVER_GROUP_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVICE_ID_FIELD_NUMBER: ClassVar[int]
    INFO_FIELD_NUMBER: ClassVar[int]
    LOGGING_FIELD_NUMBER: ClassVar[int]
    ONLINE_SERVING_ENABLED_FIELD_NUMBER: ClassVar[int]
    OPTIONS_FIELD_NUMBER: ClassVar[int]
    PREVENT_DESTROY_FIELD_NUMBER: ClassVar[int]
    REALTIME_ENVIRONMENT_FIELD_NUMBER: ClassVar[int]
    TRANSFORM_SERVER_GROUP_FIELD_NUMBER: ClassVar[int]
    VERSION_FIELD_NUMBER: ClassVar[int]
    enable_online_caching: bool
    feature_references: _containers.RepeatedCompositeFieldContainer[FeatureReference]
    feature_server_group: ServerGroupReference
    feature_service_id: _id__client_pb2.Id
    info: _basic_info__client_pb2.BasicInfo
    logging: LoggingConfigArgs
    online_serving_enabled: bool
    options: _containers.ScalarMap[str, str]
    prevent_destroy: bool
    realtime_environment: str
    transform_server_group: ServerGroupReference
    version: _framework_version__client_pb2.FrameworkVersion
    def __init__(self, feature_service_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., info: Optional[Union[_basic_info__client_pb2.BasicInfo, Mapping]] = ..., version: Optional[Union[_framework_version__client_pb2.FrameworkVersion, str]] = ..., prevent_destroy: bool = ..., options: Optional[Mapping[str, str]] = ..., enable_online_caching: bool = ..., feature_references: Optional[Iterable[Union[FeatureReference, Mapping]]] = ..., online_serving_enabled: bool = ..., logging: Optional[Union[LoggingConfigArgs, Mapping]] = ..., realtime_environment: Optional[str] = ..., transform_server_group: Optional[Union[ServerGroupReference, Mapping]] = ..., feature_server_group: Optional[Union[ServerGroupReference, Mapping]] = ...) -> None: ...

class LoggingConfigArgs(_message.Message):
    __slots__ = ["log_effective_times", "sample_rate"]
    LOG_EFFECTIVE_TIMES_FIELD_NUMBER: ClassVar[int]
    SAMPLE_RATE_FIELD_NUMBER: ClassVar[int]
    log_effective_times: bool
    sample_rate: float
    def __init__(self, sample_rate: Optional[float] = ..., log_effective_times: bool = ...) -> None: ...

class ServerGroupReference(_message.Message):
    __slots__ = ["name", "server_group_id"]
    NAME_FIELD_NUMBER: ClassVar[int]
    SERVER_GROUP_ID_FIELD_NUMBER: ClassVar[int]
    name: str
    server_group_id: _id__client_pb2.Id
    def __init__(self, server_group_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., name: Optional[str] = ...) -> None: ...



import math
from datetime import datetime
from datetime import timezone
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pandas
import pyarrow

from tecton_core import feature_set_config
from tecton_core import specs
from tecton_core.errors import TectonValidationError
from tecton_core.fco_container import FcoContainer
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.pipeline.feature_pipeline import NodeValueType
from tecton_core.pipeline.rtfv_pipeline import RealtimeFeaturePipeline
from tecton_core.realtime_context import REQUEST_TIMESTAMP_FIELD_NAME
from tecton_core.realtime_context import RealtimeContext
from tecton_core.schema import Schema
from tecton_core.schema_validation import CastError
from tecton_core.schema_validation import cast
from tecton_proto.args.pipeline__client_pb2 import Pipeline
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.common.schema__client_pb2 import Schema as SchemaProto


class PandasRealtimeFeaturePipeline(RealtimeFeaturePipeline):
    def __init__(
        self,
        name: str,
        pipeline: Pipeline,
        transformations: List[specs.TransformationSpec],
        fco_container: FcoContainer,
        id: str,
        # Input dataframe containing the necessary inputs to the Realtime Feature Pipeline
        data_df: Union[pandas.DataFrame, pyarrow.RecordBatch],
        # Only used by Tecton on Snowflake to uppercase all column names. Should be deprecated
        # with it.
        column_name_updater: Callable,
        is_prompt: bool,
        events_df_timestamp_field: Optional[str] = None,
        pipeline_inputs: Optional[Dict[str, Union[Dict[str, Any], pandas.DataFrame, RealtimeContext]]] = None,
    ) -> None:
        self._data_df = data_df
        self._fco_container = fco_container
        self._fv_id = id
        self._column_name_updater = column_name_updater
        self._context_param_name = None
        self._num_rows = (
            self._data_df.num_rows if isinstance(self._data_df, pyarrow.RecordBatch) else len(self._data_df)
        )

        # For Python Mode we go through each row of inputs in _data_df and run the pipeline for
        # each of them.
        self._current_row_index = 0
        # Cache so we don't run the Spine -> Pandas Input DF logic for every row
        self._input_name_to_df = {}

        super().__init__(
            name=name,
            pipeline=pipeline,
            transformations=transformations,
            is_prompt=is_prompt,
            events_df_timestamp_field=events_df_timestamp_field,
            pipeline_inputs=pipeline_inputs,
        )

    @classmethod
    def from_feature_definition(
        cls,
        fdw: FeatureDefinitionWrapper,
        data_df: Union[pandas.DataFrame, pyarrow.RecordBatch],
        column_name_updater: Callable,
        events_df_timestamp_field: Optional[str] = None,
    ) -> "PandasRealtimeFeaturePipeline":
        return cls(
            fdw.name,
            fdw.pipeline,
            fdw.transformations,
            fdw.fco_container,
            fdw.id,
            data_df,
            column_name_updater,
            fdw.is_prompt,
            events_df_timestamp_field,
        )

    def get_dataframe(self):
        if self.is_pandas_mode:
            return self._node_to_value(self._pipeline.root)
        # For Python Mode, we go through each row of inputs in the Pandas DF
        # self._data_df and run the Pipeline for each of them
        elif self.is_python_mode:
            rtfv_result_list = []
            for row_index in range(self._num_rows):
                self._current_row_index = row_index
                rtfv_result_dict = self._node_to_value(self._pipeline.root)
                rtfv_result_list.append(rtfv_result_dict)
            return pandas.DataFrame.from_dict(rtfv_result_list)
        else:
            msg = "Realtime Feature Views only support Pandas or Python Mode."
            raise Exception(msg)

    def _node_to_value(self, pipeline_node: PipelineNode) -> NodeValueType:
        value = super()._node_to_value(pipeline_node)

        # For Python Mode, only get the current row of inputs
        if self.is_python_mode and isinstance(value, pandas.DataFrame):
            input_dict = value.iloc[self._current_row_index].to_dict()
            return self._format_values_for_python_mode(input_dict)
        elif self.is_pandas_mode and isinstance(value, pandas.DataFrame):
            return self._format_values_for_pandas_mode(value)

        return value

    @staticmethod
    def _format_values_for_python_mode(input_dict: Dict[str, Any]) -> Dict[str, Any]:
        for key, value in input_dict.items():
            if isinstance(value, datetime):
                input_dict[key] = value.replace(tzinfo=timezone.utc)
            if value is pandas.NaT:
                input_dict[key] = None
            if isinstance(value, float) and math.isnan(value):
                input_dict[key] = None
        return input_dict

    def _request_data_node_to_value(self, pipeline_node: PipelineNode) -> Union[Dict[str, Any], pandas.DataFrame]:
        input_name = pipeline_node.request_data_source_node.input_name

        # Use cache for Python Mode
        if input_name in self._input_name_to_df:
            return self._input_name_to_df[input_name]

        request_context_schema = pipeline_node.request_data_source_node.request_context.tecton_schema

        if isinstance(self._data_df, pandas.DataFrame):
            # TODO(Oleksii): remove this path once Snowflake & Athena are deprecated
            input_df = self._get_request_context_pandas_df(request_context_schema, input_name)
        elif isinstance(self._data_df, pyarrow.RecordBatch):
            try:
                input_df = cast(self._data_df, Schema(request_context_schema)).to_pandas()
            except CastError as exc:
                msg = f"{self._fco_name} {self._name} has a dependency on the Request Data Source named '{input_name}', but it didn't pass schema validation: "
                raise CastError(msg + str(exc)) from None
        else:
            msg = f"Unexpected input dataframe type: {type(self._data_df)}"
            raise RuntimeError(msg)

        self._input_name_to_df[input_name] = input_df
        return input_df

    def _feature_view_node_to_value(self, pipeline_node: PipelineNode) -> Union[Dict[str, Any], pandas.DataFrame]:
        input_name = pipeline_node.feature_view_node.input_name

        # Use cache for Python Mode
        if input_name in self._input_name_to_df:
            return self._input_name_to_df[input_name]

        fv_features = feature_set_config.find_dependent_feature_set_items(
            self._fco_container, pipeline_node, {}, self._fv_id
        )[0]
        # Generate dependent column mappings since dependent FV have
        # internal column names with _udf_internal
        select_columns_and_rename_map = {}
        for f in fv_features.features:
            column_name = self._column_name_updater(f"{fv_features.namespace}__{f}")
            mapped_name = self._column_name_updater(f)
            select_columns_and_rename_map[column_name] = mapped_name
        if isinstance(self._data_df, pandas.DataFrame):
            # TODO(Oleksii): remove this path once Snowflake & Athena are deprecated
            feature_view_input_df = self._rename_pandas_columns(select_columns_and_rename_map, input_name)
        elif isinstance(self._data_df, pyarrow.RecordBatch):
            columns = []
            for f in select_columns_and_rename_map:
                try:
                    columns.append(self._data_df.column(f))
                except KeyError:
                    msg = f"{self._fco_name} {self._name} has a dependency on the Feature View '{input_name}'. Feature {f} of this Feature View is not found in the retrieved historical data. Available columns: {list(self._data_df.column_names)}"
                    raise TectonValidationError(msg)

            feature_view_input_df = pyarrow.RecordBatch.from_arrays(
                columns, names=list(select_columns_and_rename_map.values())
            ).to_pandas()
        else:
            msg = f"Unexpected input dataframe type: {type(self._data_df)}"
            raise RuntimeError(msg)

        self._input_name_to_df[input_name] = feature_view_input_df
        return feature_view_input_df

    def _context_node_to_value(self, pipeline_node: PipelineNode) -> Optional[RealtimeContext]:
        if not isinstance(self._data_df, pyarrow.RecordBatch):
            msg = "Realtime Context is only supported on Rift and Spark Feature Views."
            raise Exception(msg)

        data_df = self._data_df.to_pandas()
        input_name = pipeline_node.context_node.input_name

        # Use cache for Python Mode
        if input_name in self._input_name_to_df:
            context_df = self._input_name_to_df[input_name]
            context_dict = context_df.iloc[self._current_row_index].to_dict()
            return RealtimeContext(request_timestamp=context_dict[REQUEST_TIMESTAMP_FIELD_NAME], _is_python_mode=True)

        if self._events_df_timestamp_field is None or self._events_df_timestamp_field not in data_df:
            timestamp_field = self._events_df_timestamp_field if self._events_df_timestamp_field is not None else ""
            msg = f"Unable to extract timestamp field '{timestamp_field}' from events dataframe."
            raise Exception(msg)

        context_df = data_df[[self._events_df_timestamp_field]].rename(
            columns={self._events_df_timestamp_field: REQUEST_TIMESTAMP_FIELD_NAME}
        )

        self._input_name_to_df[input_name] = context_df
        if self.is_python_mode:
            context_dict = context_df.iloc[self._current_row_index].to_dict()
            return RealtimeContext(request_timestamp=context_dict[REQUEST_TIMESTAMP_FIELD_NAME], _is_python_mode=True)
        else:
            # For Pandas Mode, set the entire context dataframe
            return RealtimeContext(row_level_data=context_df, _is_python_mode=False)

    # TODO(Oleksii): remove this once Snowflake & Athena are deprecated
    def _rename_pandas_columns(
        self, select_columns_and_rename_map: Dict[str, str], input_name: str
    ) -> pandas.DataFrame:
        for f in select_columns_and_rename_map.keys():
            if f not in self._data_df.columns:
                msg = f"{self._fco_name} {self._name} has a dependency on the Feature View '{input_name}'. Feature {f} of this Feature View is not found in the retrieved historical data. Available columns: {list(self._data_df.columns)}"
                raise TectonValidationError(msg)

        # Select all of the features of the input FV from data_df
        return self._data_df.rename(columns=select_columns_and_rename_map)[[*select_columns_and_rename_map.values()]]

    # TODO(Oleksii): remove this once Snowflake & Athena are deprecated
    def _get_request_context_pandas_df(self, request_context_schema: SchemaProto, input_name: str) -> pandas.DataFrame:
        request_context_fields = [self._column_name_updater(c.name) for c in request_context_schema.columns]
        for f in request_context_fields:
            if f not in self._data_df.columns:
                msg = f"{self._fco_name} {self._name} has a dependency on the Request Data Source named '{input_name}'. Field {f} of this Request Data Source is not found in the spine. Available columns: {list(self._data_df.columns)}"
                raise TectonValidationError(msg)

        return self._data_df[request_context_fields]



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class TectonDeltaMetadata(_message.Message):
    __slots__ = ["deletion_path", "feature_start_time", "ingest_path"]
    DELETION_PATH_FIELD_NUMBER: ClassVar[int]
    FEATURE_START_TIME_FIELD_NUMBER: ClassVar[int]
    INGEST_PATH_FIELD_NUMBER: ClassVar[int]
    deletion_path: str
    feature_start_time: _timestamp_pb2.Timestamp
    ingest_path: str
    def __init__(self, deletion_path: Optional[str] = ..., ingest_path: Optional[str] = ..., feature_start_time: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ...) -> None: ...



from __future__ import annotations

import datetime
import sys
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional

import click

import tecton_core.tecton_pendulum as pendulum
from tecton import tecton_context
from tecton._internals import metadata_service
from tecton.cli import cli_utils
from tecton.cli import printer
from tecton.cli.cli_utils import cli_indent
from tecton.cli.command import TectonGroup
from tecton_core.errors import TectonNotFoundError
from tecton_core.id_helper import IdHelper
from tecton_core.specs.utils import get_timestamp_field_or_none
from tecton_proto.data import state_update__client_pb2 as state_update_pb2
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2


def _format_date(datetime: Optional[pendulum.DateTime]):
    if datetime:
        return datetime.strftime("%Y-%m-%d %H:%M:%S %Z")


@dataclass
class IntegrationTestSummaries:
    # This is a map of FeatureViewName to the list of integration test statuses for all integration tests
    #   run for that FeatureView as part of the Plan Integration Tests.
    statuses: Dict[str, List[state_update_pb2.IntegrationTestJobStatus]]

    def has_integration_tests(self):
        return bool(self.all_test_statuses())

    def all_test_statuses(self):
        all_test_statuses = []
        for _, status_list in self.statuses.items():
            all_test_statuses.extend(status_list)
        return all_test_statuses

    @staticmethod
    def _summarize_status(integration_status_list: List) -> str:
        """Given a list of integration test statuses, summarize the state of the entire bunch."""
        if not integration_status_list:
            return "No Tests"
        elif all(
            integration_status == state_update_pb2.IntegrationTestJobStatus.JOB_STATUS_SUCCEED
            for integration_status in integration_status_list
        ):
            return "Succeeded"
        elif any(
            integration_status == state_update_pb2.IntegrationTestJobStatus.JOB_STATUS_FAILED
            for integration_status in integration_status_list
        ):
            return "Failed"
        elif any(
            integration_status == state_update_pb2.IntegrationTestJobStatus.JOB_STATUS_CANCELLED
            for integration_status in integration_status_list
        ):
            return "Canceled"
        elif any(
            integration_status == state_update_pb2.IntegrationTestJobStatus.JOB_STATUS_RUNNING
            for integration_status in integration_status_list
        ):
            return "Running"
        elif any(
            integration_status == state_update_pb2.IntegrationTestJobStatus.JOB_STATUS_NOT_STARTED
            for integration_status in integration_status_list
        ):
            return "Not Started"
        else:
            return "Unknown Status"

    def summarize_status_for_all_tests(self):
        return self._summarize_status(self.all_test_statuses())

    def summarize_status_by_fv(self):
        return {fv_name: self._summarize_status(status_list) for fv_name, status_list in self.statuses.items()}

    @classmethod
    def from_protobuf(cls, successful_plan_output: state_update_pb2.SuccessfulPlanOutput):
        statuses = {}
        for test_summary in successful_plan_output.test_summaries:
            test_job_statuses = [job_summary.status for job_summary in test_summary.job_summaries]
            statuses[test_summary.feature_view_name] = test_job_statuses
        return cls(statuses=statuses)


@dataclass
class PlanListItem:
    plan_id: str
    applied_by: Optional[str]
    applied_at: Optional[pendulum.DateTime]
    created_by: str
    created_at: pendulum.DateTime
    workspace: str
    sdk_version: str
    integration_test_statuses: IntegrationTestSummaries

    @property
    def applied(self):
        if bool(self.applied_by):
            return "Applied"
        else:
            return "Created"

    @classmethod
    def from_proto(cls, state_update_entry: state_update_pb2.StateUpdateEntry):
        applied_by = cli_utils.display_principal(state_update_entry.applied_by_principal, state_update_entry.applied_by)
        applied_at = get_timestamp_field_or_none(state_update_entry, "applied_at")
        created_at = get_timestamp_field_or_none(state_update_entry, "created_at")
        return cls(
            # commit_id is called plan_id in public facing UX. Re-aliasing here.
            plan_id=state_update_entry.commit_id,
            applied_by=applied_by,
            applied_at=applied_at,
            created_by=state_update_entry.created_by,
            created_at=created_at,
            workspace=state_update_entry.workspace or "prod",
            sdk_version=state_update_entry.sdk_version,
            integration_test_statuses=IntegrationTestSummaries.from_protobuf(state_update_entry.successful_plan_output),
        )


@dataclass
class PlanSummary:
    applied_at: Optional[datetime.datetime]
    applied_by: Optional[str]
    applied: bool
    created_at: datetime.datetime
    created_by: str
    workspace: str
    sdk_version: str
    plan_url: str
    integration_test_statuses: IntegrationTestSummaries

    @classmethod
    def from_proto(cls, query_state_update_response: metadata_service_pb2.QueryStateUpdateResponseV2):
        applied_at = get_timestamp_field_or_none(query_state_update_response, "applied_at")
        applied_by = cli_utils.display_principal(
            query_state_update_response.applied_by_principal, query_state_update_response.applied_by
        )
        applied = bool(applied_at)
        created_at = get_timestamp_field_or_none(query_state_update_response, "created_at")
        return cls(
            applied=applied,
            applied_at=applied_at,
            applied_by=applied_by,
            created_at=created_at,
            created_by=query_state_update_response.created_by,
            workspace=query_state_update_response.workspace or "prod",
            sdk_version=query_state_update_response.sdk_version,
            plan_url=query_state_update_response.successful_plan_output.plan_url,
            integration_test_statuses=IntegrationTestSummaries.from_protobuf(
                query_state_update_response.successful_plan_output
            ),
        )


def get_plans_list_items(workspace: str, limit: int):
    request = metadata_service_pb2.GetStateUpdatePlanListRequest(workspace=workspace, limit=limit)
    response = metadata_service.instance().GetStateUpdatePlanList(request)
    return [PlanListItem.from_proto(entry) for entry in response.entries]


def get_plan(workspace: str, plan_id: str):
    plan_id = IdHelper.from_string(plan_id)
    request = metadata_service_pb2.QueryStateUpdateRequestV2(
        state_id=plan_id, workspace=workspace, no_color=True, json_output=False, suppress_warnings=False
    )
    try:
        response = metadata_service.instance().QueryStateUpdateV2(request)
    except TectonNotFoundError:
        printer.safe_print(
            f'Plan id "{plan_id}" not found in workspace {workspace}. Run `tecton plan-info list` to see list of '
            f"available plans."
        )
        sys.exit(1)
    return PlanSummary.from_proto(response.response_proto)


@click.group("plan-info", cls=TectonGroup)
def plan_info():
    r"""View info about plans."""


@plan_info.command(uses_workspace=True)
@click.option("--limit", default=10, type=int, help="Number of log entries to return.")
def list(limit):
    """List previous plans created for this workspace."""
    workspace = tecton_context.get_current_workspace()
    entries = get_plans_list_items(workspace, limit)
    table_rows = [
        (
            entry.plan_id,
            entry.applied,
            entry.integration_test_statuses.summarize_status_for_all_tests(),
            entry.created_by,
            _format_date(entry.created_at),
            entry.sdk_version,
        )
        for entry in entries
    ]
    cli_utils.display_table(
        ["Plan Id", "Plan Status", "Test Status", "Created by", "Creation Date", "SDK Version"], table_rows
    )


@plan_info.command()
@click.argument("plan-id", required=True, metavar="PLAN_ID")
def show(plan_id):
    """Show detailed info about a plan."""
    workspace = tecton_context.get_current_workspace()
    plan = get_plan(plan_id=plan_id, workspace=workspace)
    printer.safe_print(f"Showing current status for Plan {plan_id}")
    printer.safe_print()
    printer.safe_print(f"Plan Started At: {_format_date(plan.created_at)}")
    printer.safe_print(f"Plan Created By: {plan.created_by}")
    printer.safe_print(f"Plan Applied: {plan.applied}")
    if plan.applied:
        printer.safe_print(f"Applied At: {_format_date(plan.applied_at)}")
        printer.safe_print(f"Applied By: {plan.applied_by}")

    test_statuses = plan.integration_test_statuses
    printer.safe_print(f"Integration Test Status: {test_statuses.summarize_status_for_all_tests()}")
    if test_statuses.has_integration_tests():
        printer.safe_print("Status by Feature View:")
        for fv, status in test_statuses.summarize_status_by_fv().items():
            printer.safe_print(f"{cli_indent()}{fv}: {status}")
    printer.safe_print()
    printer.safe_print(f"View your plan in the Web UI: {plan.plan_url}")
    printer.safe_print()



import datetime
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import attrs
from typeguard import typechecked

from tecton import FeatureMetadata
from tecton import FeatureReference
from tecton import types
from tecton._internals import type_utils
from tecton.framework import base_tecton_object
from tecton.framework import configs
from tecton.framework import data_source as framework_data_source
from tecton.framework import entity as framework_entity
from tecton.framework import repo_config
from tecton.framework import utils
from tecton.framework.configs import PushConfig
from tecton.framework.data_source import BatchConfigType
from tecton.framework.data_source import BatchSource as BatchSource_v1_0
from tecton.framework.data_source import DataSource as DataSource_v1_0
from tecton.framework.data_source import FilteredSource
from tecton.framework.data_source import StreamSource as StreamSource_v1_0
from tecton.framework.entity import Entity as Entity_v1_0
from tecton.framework.feature_service import FeatureService as FeatureService_v1_0
from tecton.framework.feature_view import PIPELINE_MODE
from tecton.framework.feature_view import AggregationLeadingEdge
from tecton.framework.feature_view import BatchFeatureView as BatchFeatureView_v1_0
from tecton.framework.feature_view import BatchTriggerType
from tecton.framework.feature_view import FeatureTable as FeatureTable_v1_0
from tecton.framework.feature_view import FeatureView
from tecton.framework.feature_view import RealtimeFeatureView as RealtimeFeatureView_v1_0
from tecton.framework.feature_view import StreamFeatureView as StreamFeatureView_v1_0
from tecton.framework.feature_view import StreamProcessingMode
from tecton.framework.feature_view import infer_batch_compute_mode
from tecton.framework.transformation import Transformation as Transformation_v1_0
from tecton_core import conf
from tecton_core import feature_definition_wrapper
from tecton_core import id_helper
from tecton_core import specs
from tecton_core import time_utils
from tecton_core.compute_mode import BatchComputeMode
from tecton_core.data_types import UnknownType
from tecton_core.errors import TectonValidationError
from tecton_core.feature_definition_wrapper import FrameworkVersion
from tecton_proto.args import basic_info__client_pb2 as basic_info_pb2
from tecton_proto.args import entity__client_pb2 as entity__args_pb2
from tecton_proto.args import feature_view__client_pb2 as feature_view__args_pb2
from tecton_proto.common import schema__client_pb2 as schema_pb2
from tecton_spark import spark_schema_wrapper


class Entity(Entity_v1_0):
    """A Tecton Entity, used to organize and join features.

    An Entity is a class that represents an Entity that is being modeled in Tecton. Entities are used to index and
    organize features - a :class:`FeatureView` contains at least one Entity.

    Entities contain metadata about *join keys*, which represent the columns that are used to join features together.

    Example of an Entity declaration:

    .. code-block:: python

        from tecton import Entity

        customer = Entity(
            name='customer',
            join_keys=['customer_id'],
            description='A customer subscribing to a Sports TV subscription service',
            owner='matt@tecton.ai',
            tags={'release': 'development'}
    """

    _framework_version = FrameworkVersion.FWV5

    @typechecked
    def __init__(
        self,
        *,
        name: str,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = None,
        prevent_destroy: bool = False,
        join_keys: Optional[Union[str, List[str]]] = None,
        options: Optional[Dict[str, str]] = None,
    ):
        """Declare a new Entity.

        :param name: Unique name for the new entity.
        :param description: Short description of the new entity.
        :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
        :param owner: Owner name (typically the email of the primary maintainer).
        :param prevent_destroy: If True, this Tecton object will be blocked from being deleted or re-created (i.e. a
            destructive update) during tecton plan/apply. To remove or update this object, `prevent_destroy` must be
            set to False via the same tecton apply or a separate tecton apply. `prevent_destroy` can be used to prevent accidental changes
            such as inadvertantly deleting a Feature Service used in production or recreating a Feature View that
            triggers expensive rematerialization jobs. `prevent_destroy` also blocks changes to dependent Tecton objects
            that would trigger a recreate of the tagged object, e.g. if `prevent_destroy` is set on a Feature Service,
            that will also prevent deletions or re-creates of Feature Views used in that service. `prevent_destroy` is
            only enforced in live (i.e. non-dev) workspaces.
        :param join_keys: Names of columns that uniquely identify the entity in FeatureView's SQL statement
            for which features should be aggregated. Defaults to using ``name`` as the entity's join key.
        :param options: Additional options to configure the Entity. Used for advanced use cases and beta features.

        :raises TectonValidationError: if the input non-parameters are invalid.
        """
        from tecton_core.repo_file_handler import construct_fco_source_info

        if not join_keys:
            resolved_join_keys = schema_pb2.Column(name=name, offline_data_type=UnknownType().proto)
        elif isinstance(join_keys, str):
            resolved_join_keys = [schema_pb2.Column(name=join_keys, offline_data_type=UnknownType().proto)]
        else:
            resolved_join_keys = [
                schema_pb2.Column(name=join_key, offline_data_type=UnknownType().proto) for join_key in join_keys
            ]

        args = entity__args_pb2.EntityArgs(
            entity_id=id_helper.IdHelper.generate_id(),
            info=basic_info_pb2.BasicInfo(name=name, description=description, tags=tags, owner=owner),
            join_keys=resolved_join_keys,
            version=feature_definition_wrapper.FrameworkVersion.FWV5.value,
            prevent_destroy=prevent_destroy,
            options=options,
        )
        info = base_tecton_object.TectonObjectInfo.from_args_proto(args.info, args.entity_id)
        source_info = construct_fco_source_info(args.entity_id)
        self.__attrs_init__(info=info, spec=None, args=args, source_info=source_info)

        # Note! This is 1.0 behavior required by validation on creation
        if not conf.get_bool("TECTON_SKIP_OBJECT_VALIDATION"):
            self._validate()
        self._spec = specs.EntitySpec.from_args_proto(self._args)
        base_tecton_object._register_local_object(self)


def _get_or_create_pipeline_function(
    name: str,
    mode: str,
    description: Optional[str],
    owner: Optional[str],
    tags: Optional[Dict[str, str]],
    feature_view_function,
):
    """Helper method for creating the pipeline function from args"""
    if mode == PIPELINE_MODE:
        pipeline_function = feature_view_function
    else:
        # Use transformation from v09_compat (this file)
        inferred_transform = transformation(mode, name, description, owner, tags=tags)(feature_view_function)

        def pipeline_function(**kwargs):
            return inferred_transform(**kwargs)

    return pipeline_function


class OnDemandFeatureView(RealtimeFeatureView_v1_0):
    # TODO(FE-2269): Move deprecated function implementations to v09 objects
    _framework_version = FrameworkVersion.FWV5

    # override this function to ensure transformation associated with feature view gets the right framework version set
    @staticmethod
    def _get_or_create_pipeline_function(
        name: str,
        mode: str,
        description: Optional[str],
        owner: Optional[str],
        tags: Optional[Dict[str, str]],
        feature_view_function,
    ):
        return _get_or_create_pipeline_function(
            name=name,
            mode=mode,
            description=description,
            owner=owner,
            tags=tags,
            feature_view_function=feature_view_function,
        )

    def feature_metadata(self) -> List[FeatureMetadata]:
        msg = "v09_compat feature views do not support feature metadata"
        raise NotImplementedError(msg)


@typechecked
def on_demand_feature_view(
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    owner: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    prevent_destroy: bool = False,
    mode: str,
    sources: List[Union[configs.RequestSource, FeatureView, "FeatureReference"]],
    schema: List[types.Field],
    environments: Optional[List[str]] = None,
):
    """
    Declare an On-Demand Feature View

    :param mode: Whether the annotated function is a pipeline function ("pipeline" mode) or a transformation function ("python" or "pandas" mode).
        For the non-pipeline mode, an inferred transformation will also be registered.
    :param sources: The data source inputs to the feature view. An input can be a RequestSource or a materialized Feature View.
    :param schema: Tecton schema matching the expected output of the transformation.
    :param description: A human readable description.
    :param owner: Owner name (typically the email of the primary maintainer).
    :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
    :param name: Unique, human friendly name that identifies the FeatureView. Defaults to the function name.
    :param prevent_destroy: If True, this Tecton object will be blocked from being deleted or re-created (i.e. a
        destructive update) during tecton plan/apply. To remove or update this object, ``prevent_destroy`` must be
        set to False via the same tecton apply or a separate tecton apply. ``prevent_destroy`` can be used to prevent accidental changes
        such as inadvertantly deleting a Feature Service used in production or recreating a Feature View that
        triggers expensive rematerialization jobs. ``prevent_destroy`` also blocks changes to dependent Tecton objects
        that would trigger a recreate of the tagged object, e.g. if ``prevent_destroy`` is set on a Feature Service,
        that will also prevent deletions or re-creates of Feature Views used in that service. ``prevent_destroy`` is
        only enforced in live (i.e. non-dev) workspaces.
    :param environments: The environments in which this feature view can run. Defaults to `None`, which means
        the feature view can run in any environment. If specified, the feature view will only run in the specified
        environments. Learn more about environments at
        https://docs.tecton.ai/docs/defining-features/feature-views/on-demand-feature-view/on-demand-feature-view-environments.
    :return: An object of type :class:`tecton.OnDemandFeatureView`.

    An example declaration of an on-demand feature view using Python mode.
    With Python mode, the function sources will be dictionaries, and the function is expected to return a dictionary matching the schema from `output_schema`.
    Tecton recommends using Python mode for improved online serving performance.

    .. code-block:: python

        from tecton import RequestSource, on_demand_feature_view
        from tecton.types import Field, Float64, Int64

        # Define the request schema
        transaction_request = RequestSource(schema=[Field("amount", Float64)])

        # Define the output schema
        output_schema = [Field("transaction_amount_is_high", Int64)]

        # This On-Demand Feature View evaluates a transaction amount and declares it as "high", if it's higher than 10,000
        @on_demand_feature_view(
            sources=[transaction_request],
            mode='python',
            schema=output_schema,
            owner='matt@tecton.ai',
            tags={'release': 'production', 'prevent-destroy': 'true', 'prevent-recreate': 'true'},
            description='Whether the transaction amount is considered high (over $10000)'
        )

        def transaction_amount_is_high(transaction_request):
            result = {}
            result['transaction_amount_is_high'] = int(transaction_request['amount'] >= 10000)
            return result

    An example declaration of an on-demand feature view using Pandas mode.
    With Pandas mode, the function sources will be Pandas Dataframes, and the function is expected to return a Dataframe matching the schema from `output_schema`.

    .. code-block:: python

        from tecton import RequestSource, on_demand_feature_view
        from tecton.types import Field, Float64, Int64
        import pandas

        # Define the request schema
        transaction_request = RequestSource(schema=[Field("amount", Float64)])

        # Define the output schema
        output_schema = [Field("transaction_amount_is_high", Int64)]

        # This On-Demand Feature View evaluates a transaction amount and declares it as "high", if it's higher than 10,000
        @on_demand_feature_view(
            sources=[transaction_request],
            mode='pandas',
            schema=output_schema,
            owner='matt@tecton.ai',
            tags={'release': 'production', 'prevent-destroy': 'true', 'prevent-recreate': 'true'},
            description='Whether the transaction amount is considered high (over $10000)'
        )
        def transaction_amount_is_high(transaction_request):
            import pandas as pd

            df = pd.DataFrame()
            df['transaction_amount_is_high'] = (transaction_request['amount'] >= 10000).astype('int64')
            return df
    """

    def decorator(feature_view_function):
        return OnDemandFeatureView(
            name=name or feature_view_function.__name__,
            description=description,
            owner=owner,
            tags=tags,
            prevent_destroy=prevent_destroy,
            mode=mode,
            sources=sources,
            schema=schema,
            feature_view_function=feature_view_function,
            environments=environments,
            # Tecton 1.0 parameters.
            features=None,
            context_parameter_name=None,
        )

    return decorator


@typechecked
def batch_feature_view(
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    owner: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    prevent_destroy: bool = False,
    mode: str,
    sources: Sequence[Union[framework_data_source.BatchSource, FilteredSource]],
    entities: Sequence[framework_entity.Entity],
    aggregation_interval: Optional[datetime.timedelta] = None,
    aggregations: Optional[Sequence[configs.Aggregation]] = None,
    aggregation_secondary_key: Optional[str] = None,
    online: bool = False,
    offline: bool = False,
    ttl: Optional[datetime.timedelta] = None,
    feature_start_time: Optional[datetime.datetime] = None,
    lifetime_start_time: Optional[datetime.datetime] = None,
    manual_trigger_backfill_end_time: Optional[datetime.datetime] = None,
    batch_trigger: BatchTriggerType = BatchTriggerType.SCHEDULED,
    batch_schedule: Optional[datetime.timedelta] = None,
    online_serving_index: Optional[Sequence[str]] = None,
    batch_compute: Optional[configs.ComputeConfigTypes] = None,
    offline_store: Optional[Union[configs.OfflineStoreConfig, configs.ParquetConfig, configs.DeltaConfig]] = None,
    online_store: Optional[configs.OnlineStoreTypes] = None,
    monitor_freshness: bool = False,
    data_quality_enabled: Optional[bool] = None,
    skip_default_expectations: Optional[bool] = None,
    expected_feature_freshness: Optional[datetime.timedelta] = None,
    alert_email: Optional[str] = None,
    timestamp_field: Optional[str] = None,
    max_backfill_interval: Optional[datetime.timedelta] = None,
    max_batch_aggregation_interval: Optional[datetime.timedelta] = None,
    incremental_backfills: bool = False,
    schema: Optional[List[types.Field]] = None,
    run_transformation_validation: Optional[bool] = None,
    options: Optional[Dict[str, str]] = None,
    tecton_materialization_runtime: Optional[str] = None,
    cache_config: Optional[configs.CacheConfig] = None,
    batch_compaction_enabled: Optional[bool] = None,
    compaction_enabled: Optional[bool] = None,
    environment: Optional[str] = None,
):
    """Declare a Batch Feature View.

    :param mode: Whether the annotated function is a pipeline function ("pipeline" mode) or a transformation mode
        ("spark_sql", "pyspark", "snowflake_sql", "snowpark", "python", or "pandas"). For the non-pipeline mode, an
        inferred transformation will also be registered.
    :param sources: The data source inputs to the feature view.
    :param entities: The entities this feature view is associated with.
    :param aggregation_interval: How frequently the feature values are updated (for example, `"1h"` or `"6h"`). Only valid when using aggregations.
    :param aggregations: A list of :class:`Aggregation` structs.
    :param aggregation_secondary_key: Configures secondary key aggregates using the set column. Only valid when using aggregations.
    :param online: Whether the feature view should be materialized to the online feature store. (Default: False)
    :param offline: Whether the feature view should be materialized to the offline feature store. (Default: False)
    :param ttl: The TTL (or "look back window") for features defined by this feature view. This parameter determines how long features will live in the online store and how far to  "look back" relative to a training example's timestamp when generating offline training sets. Shorter TTLs improve performance and reduce costs.
    :param feature_start_time: When materialization for this feature view should start from. (Required if offline=true)
    :param lifetime_start_time: The start time for what data should be included in a lifetime aggregate. (Required if using lifetime windows)
    :param batch_schedule: The interval at which batch materialization should be scheduled.
    :param online_serving_index: (Advanced) Defines the set of join keys that will be indexed and queryable during online serving.
    :param batch_trigger: Defines the mechanism for initiating batch materialization jobs.
        One of ``BatchTriggerType.SCHEDULED`` or ``BatchTriggerType.MANUAL``.
        The default value is ``BatchTriggerType.SCHEDULED``, where Tecton will run materialization jobs based on the
        schedule defined by the ``batch_schedule`` parameter. If set to ``BatchTriggerType.MANUAL``, then batch
        materialization jobs must be explicitly initiated by the user through either the Tecton SDK or Airflow operator.
    :param batch_compute: Configuration for the batch materialization cluster.
    :param offline_store: Configuration for how data is written to the offline feature store.
    :param online_store: Configuration for how data is written to the online feature store.
    :param monitor_freshness: If true, enables monitoring when feature data is materialized to the online feature store.
    :param data_quality_enabled: If false, disables data quality metric computation and data quality dashboard.
    :param skip_default_expectations: If true, skips validating default expectations on the feature data.
    :param expected_feature_freshness: Threshold used to determine if recently materialized feature data is stale. Data is stale if ``now - most_recent_feature_value_timestamp > expected_feature_freshness``. For feature views using Tecton aggregations, data is stale if ``now - round_up_to_aggregation_interval(most_recent_feature_value_timestamp) > expected_feature_freshness``. Where ``round_up_to_aggregation_interval()`` rounds up the feature timestamp to the end of the ``aggregation_interval``. Value must be at least 2 times ``aggregation_interval``. If not specified, a value determined by the Tecton backend is used.
    :param alert_email: Email that alerts for this FeatureView will be sent to.
    :param description: A human readable description.
    :param owner: Owner name (typically the email of the primary maintainer).
    :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
    :param prevent_destroy: If True, this Tecton object will be blocked from being deleted or re-created (i.e. a
        destructive update) during tecton plan/apply. To remove or update this object, ``prevent_destroy`` must be set to
        False via the same tecton apply or a separate tecton apply. ``prevent_destroy`` can be used to prevent accidental changes
        such as inadvertantly deleting a Feature Service used in production or recreating a Feature View that
        triggers expensive rematerialization jobs. ``prevent_destroy`` also blocks changes to dependent Tecton objects
        that would trigger a recreate of the tagged object, e.g. if ``prevent_destroy`` is set on a Feature Service,
        that will also prevent deletions or re-creates of Feature Views used in that service. ``prevent_destroy`` is
        only enforced in live (i.e. non-dev) workspaces.
    :param timestamp_field: The column name that refers to the timestamp for records that are produced by the
        feature view. This parameter is optional if exactly one column is a Timestamp type. This parameter is
        required if using Tecton on Snowflake without Snowpark.
    :param name: Unique, human friendly name that identifies the FeatureView. Defaults to the function name.
    :param max_batch_aggregation_interval: Deprecated. Use max_backfill_interval instead, which has the exact same usage.
    :param max_backfill_interval: (Advanced) The time interval for which each backfill job will run to materialize
        feature data. This affects the number of backfill jobs that will run, which is
        (`<feature registration time>` - `feature_start_time`) / `max_backfill_interval`.
        Configuring the `max_backfill_interval` parameter appropriately will help to optimize large backfill jobs.
        If this parameter is not specified, then 10 backfill jobs will run (the default).

    :param incremental_backfills: If set to `True`, the feature view will be backfilled one interval at a time as
        if it had been updated "incrementally" since its feature_start_time. For example, if ``batch_schedule`` is 1 day
        and ``feature_start_time`` is 1 year prior to the current time, then the backfill will run 365 separate
        backfill queries to fill the historical feature data.
    :param manual_trigger_backfill_end_time: If set, Tecton will schedule backfill materialization jobs for this feature
        view up to this time. Materialization jobs after this point must be triggered manually. (This param is only valid
        to set if BatchTriggerType is MANUAL.)
    :param options: Additional options to configure the Feature View. Used for advanced use cases and beta features.
    :param schema: The output schema of the Feature View transformation. If provided and ``run_transformation_validations=True``,
        then Tecton will validate that the Feature View matches the expected schema.
    :param run_transformation_validation: If `True`, Tecton will execute the Feature View transformations during tecton plan/apply
        validation. If `False`, then Tecton will not execute the transformations during validation and ``schema`` must be
        set. Skipping query validation can be useful to speed up tecton plan/apply or for Feature Views that have issues
        with Tecton's validation (e.g. some pip dependencies). Default is True for Spark and Snowflake Feature Views and
        False for Python and Pandas Feature Views.
    :param tecton_materialization_runtime: Version of `tecton` package used by your job cluster.
    :param cache_config: Cache config for the Feature View. Including this option enables the feature server to use the cache
        when retrieving features for this feature view. Will only be respected if the feature service containing this feature
        view has `enable_online_caching` set to `True`.
    :param batch_compaction_enabled: Deprecated: Please use `compaction_enabled` instead which has the exact same usage.
     :param compaction_enabled: (Private preview) If `True`, Tecton will run a compaction job after each batch
        materialization job to write to the online store. This requires the use of Dynamo and uses the ImportTable API.
        Because each batch job overwrites the online store, a larger compute cluster may be required.
    :param environment: The custom environment in which materialization jobs will be run. Defaults to `None`, which means
        jobs will execute in the default Tecton environment.

    :return: An object of type :class:`tecton.BatchFeatureView`.

    Example BatchFeatureView declaration:

    .. code-block:: python

        from datetime import datetime
        from datetime import timedelta

        from fraud.entities import user
        from fraud.data_sources.credit_scores_batch import credit_scores_batch

        from tecton.v09_compat import batch_feature_view, Aggregation, FilteredSource

        @batch_feature_view(
            sources=[FilteredSource(credit_scores_batch)],
            entities=[user],
            mode='spark_sql',
            online=True,
            offline=True,
            feature_start_time=datetime(2020, 10, 10),
            batch_schedule=timedelta(days=1),
            ttl=timedelta(days=60),
            description="Features about the users most recent transaction in the past 60 days. Updated daily.",
            )

        def user_last_transaction_features(credit_scores_batch):
            return f'''
                SELECT
                    USER_ID,
                    TIMESTAMP,
                    AMOUNT as LAST_TRANSACTION_AMOUNT,
                    CATEGORY as LAST_TRANSACTION_CATEGORY
                FROM
                    {credit_scores_batch}
            '''

    Example BatchFeatureView declaration using aggregates:

    .. code-block:: python

        from datetime import datetime
        from datetime import timedelta

        from fraud.entities import user
        from fraud.data_sources.credit_scores_batch import credit_scores_batch

        from tecton.v09_compat import batch_feature_view, Aggregation, FilteredSource

        @batch_feature_view(
            sources=[FilteredSource(credit_scores_batch)],
            entities=[user],
            mode='spark_sql',
            online=True,
            offline=True,
            feature_start_time=datetime(2020, 10, 10),
            aggregations=[
                Aggregation(column="amount", function="mean", time_window=timedelta(days=1)),
                Aggregation(column="amount", function="mean", time_window=timedelta(days=30)),
            ],
            aggregation_interval=timedelta(days=1),
            description="Transaction amount statistics and total over a series of time windows, updated daily.",
            )

        def user_recent_transaction_aggregate_features(credit_scores_batch):
            return f'''
                SELECT
                    USER_ID,
                    AMOUNT,
                    TIMESTAMP
                FROM
                    {credit_scores_batch}
            '''
    """
    # TODO(deprecate_after=0.8): This warning can be completely removed in 0.9, so it can be deleted once the 0.8 branch is cut.
    if max_batch_aggregation_interval is not None:
        msg = "FeatureView.max_batch_aggregation_interval is deprecated and is no longer supported in 0.8. Please use max_backfill_interval instead. max_backfill_interval has the same semantics and is just a new name."
        raise ValueError(msg)

    if batch_compaction_enabled is not None and compaction_enabled is not None:
        msg = "FeatureView.batch_compaction_enabled is deprecated. Please use compaction_enabled instead. compaction_enabled has the same semantics and is just a new name."
        raise TectonValidationError(msg)

    compaction_enabled = compaction_enabled or batch_compaction_enabled or False

    def decorator(feature_view_function):
        return BatchFeatureView(
            name=name or feature_view_function.__name__,
            description=description,
            owner=owner,
            tags=tags,
            prevent_destroy=prevent_destroy,
            feature_view_function=feature_view_function,
            mode=mode,
            sources=sources,
            entities=entities,
            aggregation_interval=aggregation_interval,
            aggregations=aggregations,
            aggregation_secondary_key=aggregation_secondary_key,
            online=online,
            offline=offline,
            ttl=ttl,
            feature_start_time=feature_start_time,
            lifetime_start_time=lifetime_start_time,
            manual_trigger_backfill_end_time=manual_trigger_backfill_end_time,
            batch_trigger=batch_trigger,
            batch_schedule=batch_schedule,
            online_serving_index=online_serving_index,
            batch_compute=batch_compute,
            offline_store=offline_store,
            online_store=online_store,
            monitor_freshness=monitor_freshness,
            data_quality_enabled=data_quality_enabled,
            skip_default_expectations=skip_default_expectations,
            expected_feature_freshness=expected_feature_freshness,
            alert_email=alert_email,
            timestamp_field=timestamp_field,
            max_backfill_interval=max_backfill_interval,
            incremental_backfills=incremental_backfills,
            schema=schema,
            run_transformation_validation=run_transformation_validation,
            options=options,
            tecton_materialization_runtime=tecton_materialization_runtime,
            cache_config=cache_config,
            compaction_enabled=compaction_enabled,
            environment=environment,
            features=None,
        )

    return decorator


class BatchFeatureView(BatchFeatureView_v1_0):
    _framework_version = FrameworkVersion.FWV5

    # override this function to ensure transformation associated with feature view gets the right framework version set
    @staticmethod
    def _get_or_create_pipeline_function(
        name: str,
        mode: str,
        description: Optional[str],
        owner: Optional[str],
        tags: Optional[Dict[str, str]],
        feature_view_function,
    ):
        return _get_or_create_pipeline_function(
            name=name,
            mode=mode,
            description=description,
            owner=owner,
            tags=tags,
            feature_view_function=feature_view_function,
        )

    def feature_metadata(self) -> List[FeatureMetadata]:
        msg = "v09_compat feature views do not support feature metadata"
        raise NotImplementedError(msg)

    def _convert_data_sources_to_explicitly_filtered_objects(
        self,
        sources: Sequence[
            Union[
                framework_data_source.DataSource,
                FilteredSource,
                configs.RequestSource,
                FeatureView,
                FeatureReference,
            ]
        ],
    ) -> Sequence[Union[FilteredSource]]:
        new_sources = []
        for source in sources:
            # Invalid data sources for a materialized FV
            if isinstance(source, (FeatureReference, FeatureView, configs.RequestSource)):
                err = f"Unexpected source type {source.__class__} for Materialized Feature View"
                raise RuntimeError(err)
            elif isinstance(source, framework_data_source.DataSource):
                # Non-Filtered-Source Data Sources are always unfiltered in 0.9
                new_sources.append(source.unfiltered())
            else:
                new_sources.append(source)
        return new_sources


@typechecked
def stream_feature_view(
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    owner: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    prevent_destroy: bool = False,
    source: Union[framework_data_source.StreamSource, FilteredSource],
    entities: Sequence[framework_entity.Entity],
    mode: str,
    aggregation_interval: Optional[datetime.timedelta] = None,
    aggregations: Optional[Sequence[configs.Aggregation]] = None,
    aggregation_secondary_key: Optional[str] = None,
    stream_processing_mode: Optional[StreamProcessingMode] = None,
    online: bool = False,
    offline: bool = False,
    ttl: Optional[datetime.timedelta] = None,
    feature_start_time: Optional[datetime.datetime] = None,
    lifetime_start_time: Optional[datetime.datetime] = None,
    manual_trigger_backfill_end_time: Optional[datetime.datetime] = None,
    batch_trigger: BatchTriggerType = None,
    batch_schedule: Optional[datetime.timedelta] = None,
    online_serving_index: Optional[Sequence[str]] = None,
    batch_compute: Optional[configs.ComputeConfigTypes] = None,
    stream_compute: Optional[configs.ComputeConfigTypes] = None,
    offline_store: Optional[Union[configs.OfflineStoreConfig, configs.ParquetConfig, configs.DeltaConfig]] = None,
    online_store: Optional[configs.OnlineStoreTypes] = None,
    monitor_freshness: bool = False,
    data_quality_enabled: Optional[bool] = None,
    skip_default_expectations: Optional[bool] = None,
    expected_feature_freshness: Optional[datetime.timedelta] = None,
    alert_email: Optional[str] = None,
    timestamp_field: Optional[str] = None,
    max_backfill_interval: Optional[datetime.timedelta] = None,
    max_batch_aggregation_interval: Optional[datetime.timedelta] = None,
    output_stream: Optional[configs.OutputStream] = None,
    schema: Optional[List[types.Field]] = None,
    options: Optional[Dict[str, str]] = None,
    run_transformation_validation: Optional[bool] = None,
    tecton_materialization_runtime: Optional[str] = None,
    cache_config: Optional[configs.CacheConfig] = None,
    stream_compaction_enabled: Optional[bool] = None,
    batch_compaction_enabled: Optional[bool] = None,
    compaction_enabled: Optional[bool] = None,
    stream_tiling_enabled: Optional[bool] = None,
    environment: Optional[str] = None,
):
    """Declare a Stream Feature View.

    :param mode: Whether the annotated function is a pipeline function ("pipeline" mode) or a transformation function ("spark_sql" or "pyspark" mode).
        For the non-pipeline mode, an inferred transformation will also be registered.
    :param source: The data source input to the feature view.
    :param entities: The entities this feature view is associated with.
    :param aggregation_interval: How frequently the feature value is updated (for example, `"1h"` or `"6h"`)
    :param stream_processing_mode: Whether aggregations should be "batched" in time intervals or be updated continuously.
        Continuously aggregated features are fresher but more expensive. One of ``StreamProcessingMode.TIME_INTERVAL`` or
        ``StreamProcessingMode.CONTINUOUS``. Defaults to ``StreamProcessingMode.TIME_INTERVAL``.
    :param aggregations: A list of :class:`Aggregation` structs
    :param aggregation_secondary_key: Configures secondary key aggregates using the set column. Only valid when using aggregations.
    :param online: Whether the feature view should be materialized to the online feature store. (Default: False)
    :param offline: Whether the feature view should be materialized to the offline feature store. (Default: False)
    :param ttl: The TTL (or "look back window") for features defined by this feature view. This parameter determines how long features will live in the online store and how far to  "look back" relative to a training example's timestamp when generating offline training sets. Shorter TTLs improve performance and reduce costs.
    :param feature_start_time: When materialization for this feature view should start from. (Required if offline=true)
    :param batch_trigger: Defines the mechanism for initiating batch materialization jobs.
        One of ``BatchTriggerType.SCHEDULED`` or ``BatchTriggerType.MANUAL``.
        The default value is ``BatchTriggerType.SCHEDULED``, where Tecton will run materialization jobs based on the
        schedule defined by the ``batch_schedule`` parameter. If set to ``BatchTriggerType.MANUAL``, then batch
        materialization jobs must be explicitly initiated by the user through either the Tecton SDK or Airflow operator.
    :param batch_schedule: The interval at which batch materialization should be scheduled.
    :param online_serving_index: (Advanced) Defines the set of join keys that will be indexed and queryable during online serving.
    :param batch_compute: Batch materialization cluster configuration.
    :param stream_compute: Streaming materialization cluster configuration.
    :param offline_store: Configuration for how data is written to the offline feature store.
    :param online_store: Configuration for how data is written to the online feature store.
    :param monitor_freshness: If true, enables monitoring when feature data is materialized to the online feature store.
    :param data_quality_enabled: If false, disables data quality metric computation and data quality dashboard.
    :param skip_default_expectations: If true, skips validating default expectations on the feature data.
    :param expected_feature_freshness: Threshold used to determine if recently materialized feature data is stale. Data is stale if ``now - most_recent_feature_value_timestamp > expected_feature_freshness``. For feature views using Tecton aggregations, data is stale if ``now - round_up_to_aggregation_interval(most_recent_feature_value_timestamp) > expected_feature_freshness``. Where ``round_up_to_aggregation_interval()`` rounds up the feature timestamp to the end of the ``aggregation_interval``. Value must be at least 2 times ``aggregation_interval``. If not specified, a value determined by the Tecton backend is used.
    :param alert_email: Email that alerts for this FeatureView will be sent to.
    :param description: A human readable description.
    :param owner: Owner name (typically the email of the primary maintainer).
    :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
    :param prevent_destroy: If True, this Tecton object will be blocked from being deleted or re-created (i.e. a
        destructive update) during tecton plan/apply. To remove or update this object, ``prevent_destroy`` must be set
        to False via the same tecton apply or a separate tecton apply. ``prevent_destroy`` can be used to prevent accidental changes
        such as inadvertantly deleting a Feature Service used in production or recreating a Feature View that
        triggers expensive rematerialization jobs. ``prevent_destroy`` also blocks changes to dependent Tecton objects
        that would trigger a recreate of the tagged object, e.g. if ``prevent_destroy`` is set on a Feature Service,
        that will also prevent deletions or re-creates of Feature Views used in that service. ``prevent_destroy`` is
        only enforced in live (i.e. non-dev) workspaces.
    :param timestamp_field: The column name that refers to the timestamp for records that are produced by the
        feature view. This parameter is optional if exactly one column is a Timestamp type.
    :param name: Unique, human friendly name that identifies the FeatureView. Defaults to the function name.
    :param max_batch_aggregation_interval: Deprecated. Use max_backfill_interval instead, which has the exact same usage.
    :param max_backfill_interval: (Advanced) The time interval for which each backfill job will run to materialize
        feature data. This affects the number of backfill jobs that will run, which is
        (`<feature registration time>` - `feature_start_time`) / `max_backfill_interval`.
        Configuring the `max_backfill_interval` parameter appropriately will help to optimize large backfill jobs.
        If this parameter is not specified, then 10 backfill jobs will run (the default).
    :param output_stream: Configuration for a stream to write feature outputs to, specified as a :class:`tecton.framework.conifgs.KinesisOutputStream` or :class:`tecton.framework.configs.KafkaOutputStream`.
    :param schema: The output schema of the Feature View transformation. If provided and ``run_transformation_validations=True``,
        then Tecton will validate that the Feature View matches the expected schema.
    :param manual_trigger_backfill_end_time: If set, Tecton will schedule backfill materialization jobs for this feature view up to this time. Materialization jobs after this point must be triggered manually. (This param is only valid to set if BatchTriggerType is MANUAL.)
    :param options: Additional options to configure the Feature View. Used for advanced use cases and beta features.
    :param run_transformation_validation: If `True`, Tecton will execute the Feature View transformations during tecton plan/apply
        validation. If `False`, then Tecton will not execute the transformations during validation and ``schema`` must be
        set. Skipping query validation can be useful to speed up tecton plan/apply or for Feature Views that have issues
        with Tecton's validation (e.g. some pip dependencies). Default is True for Spark and Snowflake Feature Views and
        False for Python and Pandas Feature Views or Feature Views with Push Sources.
    :param tecton_materialization_runtime: Version of `tecton` package used by your job cluster.
    :param cache_config: Cache config for the Feature View. Including this option enables the feature server to use the cache
        when retrieving features for this feature view. Will only be respected if the feature service containing this feature
        view has `enable_online_caching` set to `True`.
    :param stream_compaction_enabled: Deprecated: Please use `stream_tiling_enabled` instead which has the exact same usage.
    :param batch_compaction_enabled: Deprecated: Please use `compaction_enabled` instead which has the exact same usage.
    :param stream_tiling_enabled: (Private preview) If `False`, Tecton transforms and writes all events from the stream to the online store (same as stream_processing_mode=``StreamProcessingMode.CONTINUOUS``) . If `True`, Tecton will store the partial aggregations of the events in the online store. Defaults to ``False``.
    :param compaction_enabled: (Private preview) If `True`, Tecton will run a compaction job after each batch
        materialization job to write to the online store. This requires the use of Dynamo and uses the ImportTable API.
        Becuase each batch job overwrites the online store, a larger compute cluster may be required. This is required to be True if `stream_compaction_enabled` is True. Defaults to ``False``.
    :param environment: The custom environment in which materialization jobs will be run. Defaults to `None`, which means
        jobs will execute in the default Tecton environment.

    :return: An object of type :class:`tecton.StreamFeatureView`.

    Example `StreamFeatureView` declaration:

    .. code-block:: python

        from datetime import datetime, timedelta
        from entities import user
        from transactions_stream import transactions_stream
        from tecton import Aggregation, FilteredSource, stream_feature_view

        @stream_feature_view(
            source=FilteredSource(transactions_stream),
            entities=[user],
            mode="spark_sql",
            ttl=timedelta(days=30),
            online=True,
            offline=True,
            batch_schedule=timedelta(days=1),
            feature_start_time=datetime(2020, 10, 10),
            tags={"release": "production"},
            owner="kevin@tecton.ai",
            description="Features about the users most recent transaction in the past 60 days. Updated continuously.",
        )
        def user_last_transaction_features(transactions_stream):
            return f'''
                SELECT
                    USER_ID,
                    TIMESTAMP,
                    AMOUNT as LAST_TRANSACTION_AMOUNT,
                    CATEGORY as LAST_TRANSACTION_CATEGORY
                FROM
                    {transactions_stream}
                '''

    Example `StreamFeatureView` declaration using aggregates:

    .. code-block:: python

        from datetime import datetime, timedelta
        from entities import user
        from transactions_stream import transactions_stream
        from tecton import Aggregation, FilteredSource, stream_feature_view

        @stream_feature_view(
            source=FilteredSource(transactions_stream),
            entities=[user],
            mode="spark_sql",
            aggregation_interval=timedelta(minutes=10),
            aggregations=[
                Aggregation(column="AMOUNT", function="mean", time_window=timedelta(hours=1)),
                Aggregation(column="AMOUNT", function="mean", time_window=timedelta(hours=24)),
                Aggregation(column="AMOUNT", function="mean", time_window=timedelta(hours=72)),
                Aggregation(column="AMOUNT", function="sum", time_window=timedelta(hours=1)),
                Aggregation(column="AMOUNT", function="sum", time_window=timedelta(hours=24)),
                Aggregation(column="AMOUNT", function="sum", time_window=timedelta(hours=72)),
            ],
            online=True,
            feature_start_time=datetime(2020, 10, 10),
            tags={"release": "production"},
            owner="kevin@tecton.ai",
            description="Transaction amount statistics and total over a series of time windows, updated every ten minutes.",
        )
        def user_recent_transaction_aggregate_features(transactions_stream):
            return f'''
                SELECT
                    USER_ID,
                    AMOUNT,
                    TIMESTAMP
                FROM
                    {transactions_stream}
                '''
    """
    # TODO(deprecate_after=0.8): This warning can be completely removed in 0.9, so it can be deleted once the 0.8 branch is cut.
    if max_batch_aggregation_interval is not None:
        msg = "FeatureView.max_batch_aggregation_interval is deprecated and is no longer supported in 0.8. Please use max_backfill_interval instead. max_backfill_interval has the same semantics and is just a new name."
        raise ValueError(msg)

    if batch_compaction_enabled is not None and compaction_enabled is not None:
        msg = "FeatureView.batch_compaction_enabled is deprecated. Please use compaction_enabled instead. compaction_enabled has the same semantics and is just a new name."
        raise TectonValidationError(msg)

    compaction_enabled = compaction_enabled or batch_compaction_enabled or False

    if stream_compaction_enabled is not None and stream_tiling_enabled is not None:
        msg = "FeatureView.stream_compaction_enabled is deprecated. Please use stream_tiling_enabled instead. stream_tiling_enabled has the same semantics and is just a new name."
        raise TectonValidationError(msg)

    stream_tiling_enabled = stream_tiling_enabled or stream_compaction_enabled or False

    def decorator(feature_view_function):
        return StreamFeatureView(
            name=name or feature_view_function.__name__,
            description=description,
            owner=owner,
            tags=tags,
            prevent_destroy=prevent_destroy,
            feature_view_function=feature_view_function,
            source=source,
            entities=entities,
            mode=mode,
            aggregation_interval=aggregation_interval,
            aggregations=aggregations,
            aggregation_secondary_key=aggregation_secondary_key,
            stream_processing_mode=stream_processing_mode,
            online=online,
            offline=offline,
            ttl=ttl,
            feature_start_time=feature_start_time,
            lifetime_start_time=lifetime_start_time,
            manual_trigger_backfill_end_time=manual_trigger_backfill_end_time,
            batch_trigger=batch_trigger,
            batch_schedule=batch_schedule,
            online_serving_index=online_serving_index,
            batch_compute=batch_compute,
            stream_compute=stream_compute,
            offline_store=offline_store,
            online_store=online_store,
            monitor_freshness=monitor_freshness,
            data_quality_enabled=data_quality_enabled,
            skip_default_expectations=skip_default_expectations,
            expected_feature_freshness=expected_feature_freshness,
            alert_email=alert_email,
            timestamp_field=timestamp_field,
            max_backfill_interval=max_backfill_interval,
            output_stream=output_stream,
            schema=schema,
            options=options,
            run_transformation_validation=run_transformation_validation,
            tecton_materialization_runtime=tecton_materialization_runtime,
            cache_config=cache_config,
            compaction_enabled=compaction_enabled,
            stream_tiling_enabled=stream_tiling_enabled,
            environment=environment,
            aggregation_leading_edge=AggregationLeadingEdge.LATEST_EVENT_TIME,  # new default for v09_compat objects
        )

    return decorator


class StreamFeatureView(StreamFeatureView_v1_0):
    _framework_version = FrameworkVersion.FWV5

    def feature_metadata(self) -> List[FeatureMetadata]:
        msg = "v09_compat feature views do not support feature metadata"
        raise NotImplementedError(msg)

    # override this function to ensure transformation associated with feature view gets the right framework version set
    @staticmethod
    def _get_or_create_pipeline_function(
        name: str,
        mode: str,
        description: Optional[str],
        owner: Optional[str],
        tags: Optional[Dict[str, str]],
        feature_view_function,
    ):
        return _get_or_create_pipeline_function(
            name=name,
            mode=mode,
            description=description,
            owner=owner,
            tags=tags,
            feature_view_function=feature_view_function,
        )

    def _convert_data_sources_to_explicitly_filtered_objects(
        self,
        sources: Sequence[
            Union[
                framework_data_source.DataSource,
                FilteredSource,
                configs.RequestSource,
                FeatureView,
                FeatureReference,
            ]
        ],
    ) -> Sequence[Union[FilteredSource]]:
        new_sources = []
        for source in sources:
            # Invalid data sources for a materialized FV
            if isinstance(source, (FeatureReference, FeatureView, configs.RequestSource)):
                err = f"Unexpected source type {source.__class__} for Materialized Feature View"
                raise RuntimeError(err)
            elif isinstance(source, framework_data_source.DataSource):
                # Non-Filtered-Source Data Sources are always unfiltered in 0.9
                new_sources.append(source.unfiltered())
            else:
                new_sources.append(source)
        return new_sources


@attrs.define(eq=False)
class FeatureTable(FeatureTable_v1_0):
    """A Tecton Feature Table.

    Feature Tables are used to batch push features into Tecton from external feature computation systems.

    An example declaration of a FeatureTable:

    .. code-block:: python

        from tecton import Entity, FeatureTable
        from tecton.types import Field, String, Timestamp, Int64
        import datetime

        # Declare your user Entity instance here or import it if defined elsewhere in
        # your Tecton repo.

        user = ...

        schema = [
            Field('user_id', String),
            Field('timestamp', Timestamp),
            Field('user_login_count_7d', Int64),
            Field('user_login_count_30d', Int64)
        ]

        user_login_counts = FeatureTable(
            name='user_login_counts',
            entities=[user],
            schema=schema,
            online=True,
            offline=True,
            ttl=datetime.timedelta(days=30)
        )

    Attributes:
        entities: The Entities for this Feature View.
        info: A dataclass containing basic info about this Tecton Object.
    """

    entities: Tuple[framework_entity.Entity, ...] = attrs.field(
        repr=utils.short_tecton_objects_repr, on_setattr=attrs.setters.frozen
    )

    _framework_version = FrameworkVersion.FWV5

    @typechecked
    def __init__(
        self,
        *,
        name: str,
        description: Optional[str] = None,
        owner: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        prevent_destroy: bool = False,
        entities: List[framework_entity.Entity],
        schema: List[types.Field],
        ttl: Optional[datetime.timedelta] = None,
        online: bool = False,
        offline: bool = False,
        offline_store: Optional[Union[configs.OfflineStoreConfig, configs.DeltaConfig]] = None,
        online_store: Optional[configs.OnlineStoreTypes] = None,
        batch_compute: Optional[configs.ComputeConfigTypes] = None,
        online_serving_index: Optional[List[str]] = None,
        alert_email: Optional[str] = None,
        tecton_materialization_runtime: Optional[str] = None,
    ):
        """Instantiate a new FeatureTable.

        :param name: Unique, human friendly name that identifies the FeatureTable.
        :param description: A human readable description.
        :param owner: Owner name (typically the email of the primary maintainer).
        :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
        :param prevent_destroy: If True, this Tecton object will be blocked from being deleted or re-created (i.e. a
            destructive update) during tecton plan/apply. To remove or update this object, ``prevent_destroy`` must
            be set to False via the same tecton apply or a separate tecton apply. ``prevent_destroy`` can be used to prevent accidental changes
            such as inadvertantly deleting a Feature Service used in production or recreating a Feature View that
            triggers expensive rematerialization jobs. ``prevent_destroy`` also blocks changes to dependent Tecton objects
            that would trigger a recreate of the tagged object, e.g. if ``prevent_destroy`` is set on a Feature Service,
            that will also prevent deletions or re-creates of Feature Views used in that service. ``prevent_destroy`` is
            only enforced in live (i.e. non-dev) workspaces.
        :param entities: A list of Entity objects, used to organize features.
        :param schema: A schema for the FeatureTable. Supported types are: Int64, Float64, String, Bool and Array with Int64, Float32, Float64 and String typed elements. Additionally you must have exactly one Timestamp typed column for the feature timestamp.
        :param ttl: The TTL (or "look back window") for features defined by this feature table. This parameter determines how long features will live in the online store and how far to  "look back" relative to a training example's timestamp when generating offline training sets. Shorter TTLs improve performance and reduce costs.
        :param online: Enable writing to online feature store. (Default: False)
        :param offline: Enable writing to offline feature store. (Default: False)
        :param offline_store: Configuration for how data is written to the offline feature store.
        :param online_store: Configuration for how data is written to the online feature store.
        :param batch_compute: Configuration for batch materialization clusters. Should be one of:
            [:class:`EMRClusterConfig`, :class:`DatabricksClusterConfig`, :class:`EMRJsonClusterConfig`, :class:`DatabricksJsonClusterConfig`, :class:`DataprocJsonClusterConfig`]
        :param online_serving_index: (Advanced) Defines the set of join keys that will be indexed and queryable during online serving.
            Defaults to the complete set of join keys. Up to one join key may be omitted. If one key is omitted, online requests to a Feature Service will
            return all feature vectors that match the specified join keys.
        :param alert_email: Email that alerts for this FeatureTable will be sent to.
        :param tecton_materialization_runtime: Version of `tecton` package used by your job cluster.
        """
        from tecton_core.repo_file_handler import construct_fco_source_info

        if online_store is None:
            online_store = repo_config.get_feature_table_defaults().online_store

        if offline_store is None:
            offline_store = repo_config.get_feature_table_defaults().offline_store_config
        elif isinstance(offline_store, configs.DeltaConfig):
            offline_store = configs.OfflineStoreConfig(staging_table_format=offline_store)

        if batch_compute is None:
            batch_compute = repo_config.get_feature_table_defaults().batch_compute

        if tecton_materialization_runtime is None:
            tecton_materialization_runtime = repo_config.get_feature_table_defaults().tecton_materialization_runtime

        if isinstance(schema, list):
            wrapper = type_utils.to_spark_schema_wrapper(schema)
        else:
            wrapper = spark_schema_wrapper.SparkSchemaWrapper(schema)

        feature_table_args = feature_view__args_pb2.FeatureTableArgs(
            tecton_materialization_runtime=tecton_materialization_runtime,
            schema=wrapper.to_proto(),
            serving_ttl=time_utils.timedelta_to_proto(ttl),
            batch_compute=batch_compute._to_cluster_proto(),
            online_store=online_store._to_proto() if online_store else None,
            monitoring=configs.MonitoringConfig(monitor_freshness=False, alert_email=alert_email)._to_proto(),
            offline_store=offline_store._to_proto(),
        )

        # If unspecified, online_serving_index defaults to the join_keys of the Feature Table.
        join_keys = []
        for entity in entities:
            join_keys.extend(entity.join_keys)

        args = feature_view__args_pb2.FeatureViewArgs(
            feature_table_args=feature_table_args,
            feature_view_id=id_helper.IdHelper.generate_id(),
            info=basic_info_pb2.BasicInfo(name=name, description=description, tags=tags, owner=owner),
            prevent_destroy=prevent_destroy,
            feature_view_type=feature_view__args_pb2.FeatureViewType.FEATURE_VIEW_TYPE_FEATURE_TABLE,
            version=feature_definition_wrapper.FrameworkVersion.FWV5.value,
            entities=[
                feature_view__args_pb2.EntityKeyOverride(entity_id=entity.info._id_proto, join_keys=entity.join_keys)
                for entity in entities
            ],
            online_enabled=online,
            offline_enabled=offline,
            online_serving_index=online_serving_index,
            batch_compute_mode=BatchComputeMode(
                infer_batch_compute_mode(
                    pipeline_root=None, batch_compute_config=batch_compute, stream_compute_config=None
                )
            ).value,
        )

        info = base_tecton_object.TectonObjectInfo.from_args_proto(args.info, args.feature_view_id)
        source_info = construct_fco_source_info(args.feature_view_id)
        self.__attrs_init__(
            info=info,
            feature_definition=None,
            args=args,
            source_info=source_info,
            entities=tuple(entities),
            args_supplement=None,
        )

        # Note! This is 1.0 behavior required by validation on creation
        schemas = self._get_schemas()
        self._args_supplement = self._get_args_supplement(schemas)
        if not conf.get_bool("TECTON_SKIP_OBJECT_VALIDATION"):
            self._validate()
        self._populate_spec()
        base_tecton_object._register_local_object(self)

    def feature_metadata(self) -> List[FeatureMetadata]:
        msg = "v09_compat feature views do not support feature metadata"
        raise NotImplementedError(msg)


class DataSource(DataSource_v1_0):
    _framework_version = FrameworkVersion.FWV5


class BatchSource(BatchSource_v1_0):
    _framework_version = FrameworkVersion.FWV5


class StreamSource(StreamSource_v1_0):
    _framework_version = FrameworkVersion.FWV5


class Transformation(Transformation_v1_0):
    _framework_version = FrameworkVersion.FWV5


@typechecked
def transformation(
    mode: str,
    name: Optional[str] = None,
    description: Optional[str] = None,
    owner: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    prevent_destroy: bool = False,
    options: Optional[Dict[str, str]] = None,
):
    """Declares a Transformation that wraps a user function. Transformations are assembled in a pipeline function of a Feature View.

    :param mode: The mode for this transformation must be one of "spark_sql", "pyspark", "snowflake_sql", "snowpark", "pandas" or "python".
    :param name: Unique, human friendly name that identifies the Transformation. Defaults to the function name.
    :param description: A human readable description.
    :param owner: Owner name (typically the email of the primary maintainer).
    :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
    :param prevent_destroy: If True, this Tecton object will be blocked from being deleted or re-created (i.e. a
        destructive update) during tecton plan/apply. To remove or update this object, `prevent_destroy` must be set to
        False via the same tecton apply or a separate tecton apply. `prevent_destroy` can be used to prevent accidental changes
        such as inadvertantly deleting a Feature Service used in production or recreating a Feature View that
        triggers expensive rematerialization jobs. `prevent_destroy` also blocks changes to dependent Tecton objects
        that would trigger a recreate of the tagged object, e.g. if `prevent_destroy` is set on a Feature Service,
        that will also prevent deletions or re-creates of Feature Views used in that service. `prevent_destroy` is
        only enforced in live (i.e. non-dev) workspaces.
    :param options: Additional options to configure the Transformation. Used for advanced use cases and beta features.
    :return: A wrapped transformation

    Examples of Spark SQL, PySpark, Pandas, and Python transformation declarations:

        .. code-block:: python

            from tecton import transformation
            from pyspark.sql import DataFrame
            import pandas as pd

            # Create a Spark SQL transformation.
            @transformation(mode="spark_sql",
                            description="Create new column by splitting the string in an existing column")
            def str_split(input_data, column_to_split, new_column_name, delimiter):
                return f'''
                    SELECT
                        *,
                        split({column_to_split}, {delimiter}) AS {new_column_name}
                    FROM {input_data}
                '''

            # Create a PySpark transformation.
            @transformation(mode="pyspark",
                            description="Add a new column 'user_has_good_credit' if score is > 670")
            def user_has_good_credit_transformation(credit_scores):
                from pyspark.sql import functions as F

                (df = credit_scores.withColumn("user_has_good_credit",
                    F.when(credit_scores["credit_score"] > 670, 1).otherwise(0))
                return df.select("user_id", df["date"].alias("timestamp"), "user_has_good_credit") )

            # Create a Pandas transformation.
            @transformation(mode="pandas",
                            description="Whether the transaction amount is considered high (over $10000)")
            def transaction_amount_is_high(transaction_request):
                import pandas as pd

                df = pd.DataFrame()
                df['amount_is_high'] = (request['amount'] >= 10000).astype('int64')
                return df

            @transformation(mode="python",
                            description="Whether the transaction amount is considered high (over $10000)")
            # Create a Python transformation.
            def transaction_amount_is_high(transaction_request):

                result = {}
                result['transaction_amount_is_high'] = int(transaction_request['amount'] >= 10000)
                return result
    """

    def decorator(user_function):
        transform_name = name or user_function.__name__
        transform = Transformation(
            name=transform_name,
            description=description,
            owner=owner,
            tags=tags,
            prevent_destroy=prevent_destroy,
            mode=mode,
            user_function=user_function,
            options=options,
        )

        return transform

    return decorator


class FeatureService(FeatureService_v1_0):
    _framework_version = FrameworkVersion.FWV5

    def feature_metadata(self) -> List[FeatureMetadata]:
        msg = "v09_compat feature views do not support feature metadata"
        raise NotImplementedError(msg)


class PushSource(StreamSource):
    def __init__(
        self,
        *,
        name: str,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = None,
        prevent_destroy: bool = False,
        schema: List[types.Field],
        batch_config: Optional[BatchConfigType] = None,
        options: Optional[Dict[str, str]] = None,
    ):
        super().__init__(
            name=name,
            description=description,
            tags=tags,
            owner=owner,
            prevent_destroy=prevent_destroy,
            batch_config=batch_config,
            stream_config=PushConfig(),
            options=options,
            schema=schema,
        )



from tecton_proto.spark_common import clusters__client_pb2 as _clusters__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class Library(_message.Message):
    __slots__ = ["egg", "egg_resource", "jar", "jar_resource", "maven", "pypi", "whl", "whl_resource"]
    EGG_FIELD_NUMBER: ClassVar[int]
    EGG_RESOURCE_FIELD_NUMBER: ClassVar[int]
    JAR_FIELD_NUMBER: ClassVar[int]
    JAR_RESOURCE_FIELD_NUMBER: ClassVar[int]
    MAVEN_FIELD_NUMBER: ClassVar[int]
    PYPI_FIELD_NUMBER: ClassVar[int]
    WHL_FIELD_NUMBER: ClassVar[int]
    WHL_RESOURCE_FIELD_NUMBER: ClassVar[int]
    egg: str
    egg_resource: _clusters__client_pb2.ResourceLocation
    jar: str
    jar_resource: _clusters__client_pb2.ResourceLocation
    maven: MavenLibrary
    pypi: PyPiLibrary
    whl: str
    whl_resource: _clusters__client_pb2.ResourceLocation
    def __init__(self, jar: Optional[str] = ..., egg: Optional[str] = ..., whl: Optional[str] = ..., jar_resource: Optional[Union[_clusters__client_pb2.ResourceLocation, Mapping]] = ..., egg_resource: Optional[Union[_clusters__client_pb2.ResourceLocation, Mapping]] = ..., whl_resource: Optional[Union[_clusters__client_pb2.ResourceLocation, Mapping]] = ..., maven: Optional[Union[MavenLibrary, Mapping]] = ..., pypi: Optional[Union[PyPiLibrary, Mapping]] = ...) -> None: ...

class MavenLibrary(_message.Message):
    __slots__ = ["coordinates", "exclusions", "repo"]
    COORDINATES_FIELD_NUMBER: ClassVar[int]
    EXCLUSIONS_FIELD_NUMBER: ClassVar[int]
    REPO_FIELD_NUMBER: ClassVar[int]
    coordinates: str
    exclusions: _containers.RepeatedScalarFieldContainer[str]
    repo: str
    def __init__(self, coordinates: Optional[str] = ..., repo: Optional[str] = ..., exclusions: Optional[Iterable[str]] = ...) -> None: ...

class PyPiLibrary(_message.Message):
    __slots__ = ["package", "repo"]
    PACKAGE_FIELD_NUMBER: ClassVar[int]
    REPO_FIELD_NUMBER: ClassVar[int]
    package: str
    repo: str
    def __init__(self, package: Optional[str] = ..., repo: Optional[str] = ...) -> None: ...
